"""
Security Network - Service Access Point (SN-SAP) module

This module implements the SN-SAP interface for the Security Network.

This module is based in ETSI TS 102 723-8 V1.1.1 (2016-04)
"""

from __future__ import annotations
from collections.abc import Callable
from enum import Enum


class SNSIGNRequest:
    """
    SN-SIGN.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.1.2

    ...

    Attributes
    ----------
    tbs_message_length : int
        Length of the message to be signed (16 bits range)
    tbs_message : bytes
        Message to be signed
    its_aid : int
        ITS AID (Determines the security profile to apply)
    permissions_length : int
        Length of the permissions field (16 bits range)
    permissions : bytes
        Specify the sender's permissions for the security entity to decide which key to use. (Max length 31 octets)
    context_information : bytes (optional)
        Context information which could be used in selecting properties of the underlying security protocol for various purposes.
    key_handle : int (optional)
        An indicator for the security entity to decide which key to use
    """

    def __init__(
        self,
        tbs_message_length: int,
        tbs_message: bytes,
        its_aid: int,
        permissions_length: int,
        permissions: bytes,
        context_information: bytes = None,
        key_handle: int = None,
    ):
        self.tbs_message_length: int = tbs_message_length
        self.tbs_message: bytes = tbs_message
        self.its_aid: int = its_aid
        self.permissions_length: int = permissions_length
        self.permissions: bytes = permissions
        self.context_information: bytes = context_information
        self.key_handle: int = key_handle

    def __repr__(self):
        return (
            f"SNSIGNRequest(tbs_message_length={self.tbs_message_length}, "
            f"tbs_message={self.tbs_message}, its_aid={self.its_aid}, "
            f"permissions_length={self.permissions_length}, permissions={self.permissions}, "
            f"context_information={self.context_information}, key_handle={self.key_handle})"
        )

    def __str__(self):
        return (
            f"SNSIGNRequest(tbs_message_length={self.tbs_message_length}, "
            f"tbs_message={self.tbs_message}, its_aid={self.its_aid}, "
            f"permissions_length={self.permissions_length}, permissions={self.permissions}, "
            f"context_information={self.context_information}, key_handle={self.key_handle})"
        )


class SNSIGNConfirm:
    """
    SN-SIGN.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.1.3

    ...
    Attributes
    ----------
    sec_message_length : int
        Length of the signed message (16 bits range)
    sec_message : bytes
        Signed message
    """

    def __init__(self, sec_message_length: int, sec_message: bytes):
        self.sec_message_length: int = sec_message_length
        self.sec_message: bytes = sec_message

    def __repr__(self):
        return f"SNSIGNConfirm(sec_message_length={self.sec_message_length}, sec_message={self.sec_message})"

    def __str__(self):
        return f"SNSIGNConfirm(sec_message_length={self.sec_message_length}, sec_message={self.sec_message})"


class SNVERIFYRequest:
    """
    SN-VERIFY.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.2.2

    ...
    Attributes
    ----------
    sec_header_length : int
        Length of the security header (16 bits range)
    sec_header : bytes
        Security header
    message_length : int
        Length of the message to be verified (16 bits range)
    message : bytes
        Message to be verified
    """

    def __init__(
        self,
        sec_header_length: int,
        sec_header: bytes,
        message_length: int,
        message: bytes,
    ):
        self.sec_header_length: int = sec_header_length
        self.sec_header: bytes = sec_header
        self.message_length: int = message_length
        self.message: bytes = message

    def __repr__(self):
        return (
            f"SNVERIFYRequest(sec_header_length={self.sec_header_length}, "
            f"sec_header={self.sec_header}, message_length={self.message_length}, "
            f"message={self.message})"
        )

    def __str__(self):
        return (
            f"SNVERIFYRequest(sec_header_length={self.sec_header_length}, "
            f"sec_header={self.sec_header}, message_length={self.message_length}, "
            f"message={self.message})"
        )


class ReportVerify(Enum):
    """
    ReportVerify class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.2.3 (Table 5) 2nd Row

    ...
    Attributes
    ----------
    SUCCESS
    FALSE_SIGNATURE
    INVALID_CERTIFICATE
    REVOKED_CERTIFICATE
    INCONSISTENT_CHAIN
    INVALID_TIMESTAMP
    DUPLICATE_MESSAGE
    INVALID_MOBILITY_DATA
    UNSIGNED_MESSAGE
    SIGNER_CERTIFICATE_NOT_FOUND
    UNSUPPORTED_SIGNER_IDENTIFIER_TYPE
    INCOMPATIBLE_PROTOCOL
    """

    SUCCESS = 0
    FALSE_SIGNATURE = 1
    INVALID_CERTIFICATE = 2
    REVOKED_CERTIFICATE = 3
    INCONSISTENT_CHAIN = 4
    INVALID_TIMESTAMP = 5
    DUPLICATE_MESSAGE = 6
    INVALID_MOBILITY_DATA = 7
    UNSIGNED_MESSAGE = 8
    SIGNER_CERTIFICATE_NOT_FOUND = 9
    UNSUPPORTED_SIGNER_IDENTIFIER_TYPE = 10
    INCOMPATIBLE_PROTOCOL = 11


class SNVERIFYConfirm:
    """
    SN-VERIFY.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.2.3

    ...
    Attributes
    ----------
    report : ReportVerify
        Verification report
    certificate_id : bytes
        Identification of the source certificate (8 octets)
    its_aid_length : int
        Length of the ITS AID (16 bits range)
    its_aid : bytes
        ITS AID
    permissions : bytes
        Permissions of the signer (Max length 31 octets)
    """

    def __init__(
        self,
        report: ReportVerify,
        certificate_id: bytes,
        its_aid_length: int,
        its_aid: bytes,
        permissions: bytes,
    ):
        self.report: ReportVerify = report
        self.certificate_id: bytes = certificate_id
        self.its_aid_length: int = its_aid_length
        self.its_aid: bytes = its_aid
        self.permissions: bytes = permissions

    def __repr__(self):
        return (
            f"SNVERIFYConfirm(report={self.report}, certificate_id={self.certificate_id}, "
            f"its_aid_length={self.its_aid_length}, its_aid={self.its_aid}, "
            f"permissions={self.permissions})"
        )

    def __str__(self):
        return (
            f"SNVERIFYConfirm(report={self.report}, certificate_id={self.certificate_id}, "
            f"its_aid_length={self.its_aid_length}, its_aid={self.its_aid}, "
            f"permissions={self.permissions})"
        )


class SNENCRYPTRequest:
    """
    SN-ENCRYPT.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.3.2

    ...
    Attributes
    ----------
    tbe_payload_length : int
        Length of the payload to be encrypted (16 bits range)
    tbe_payload : bytes
        Payload to be encrypted
    target_id_list_length : int
        Length of the list of target identifiers (16 bits range)
    target_id_list : list[bytes]
        List of target identifiers
    context_information : bytes (optional)
        Context information
    """

    def __init__(
        self,
        tbe_payload_length: int,
        tbe_payload: bytes,
        target_id_list_length: int,
        target_id_list: list[bytes],
        context_information: bytes = None,
    ):
        self.tbe_payload_length: int = tbe_payload_length
        self.tbe_payload: bytes = tbe_payload
        self.target_id_list_length: int = target_id_list_length
        self.target_id_list: list[bytes] = target_id_list
        self.context_information: bytes = context_information

    def __repr__(self):
        return (
            f"SNENCRYPTRequest(tbe_payload_length={self.tbe_payload_length}, tbe_payload={self.tbe_payload}, "
            f"target_id_list_length={self.target_id_list_length}, target_id_list={self.target_id_list}, "
            f"context_information={self.context_information})"
        )

    def __str__(self):
        return (
            f"SNENCRYPTRequest(tbe_payload_length={self.tbe_payload_length}, tbe_payload={self.tbe_payload}, "
            f"target_id_list_length={self.target_id_list_length}, target_id_list={self.target_id_list}, "
            f"context_information={self.context_information})"
        )


class SNENCRYPTConfirm:
    """
    SN-ENCRYPT.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.3.3

    ...
    Attributes
    ----------
    encrypted_message_length : int
        Length of the encrypted message (16 bits range)
    encrypted_message : bytes
        Encrypted message
    """

    def __init__(self, encrypted_message_length: int, encrypted_message: bytes):
        self.encrypted_message_length: int = encrypted_message_length
        self.encrypted_message: bytes = encrypted_message

    def __repr__(self):
        return (
            f"SNENCRYPTConfirm(encrypted_message_length={self.encrypted_message_length}, "
            f"encrypted_message={self.encrypted_message})"
        )

    def __str__(self):
        return (
            f"SNENCRYPTConfirm(encrypted_message_length={self.encrypted_message_length}, "
            f"encrypted_message={self.encrypted_message})"
        )


class SNDECRYPTRequest:
    """
    SN-DECRYPT.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.4.2

    ...
    Attributes
    ----------
    encrypted_message_length : int
        Length of the encrypted message (16 bits range)
    encrypted_message : bytes
        Encrypted message
    """

    def __init__(self, encrypted_message_length: int, encrypted_message: bytes):
        self.encrypted_message_length: int = encrypted_message_length
        self.encrypted_message: bytes = encrypted_message

    def __repr__(self):
        return (
            f"SNDECRYPTRequest(encrypted_message_length={self.encrypted_message_length}, "
            f"encrypted_message={self.encrypted_message})"
        )

    def __str__(self):
        return (
            f"SNDECRYPTRequest(encrypted_message_length={self.encrypted_message_length}, "
            f"encrypted_message={self.encrypted_message})"
        )


class ReportDecrypt(Enum):
    """
    ReportDecrypt class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.4.3 Table 9 Row 4

    ...
    Attributes
    ----------
    SUCCESS
    UNENCRYPTED_MESSAGE
    DECRYPTION_ERROR
    INCOMPATIBLE_PROTOCOL
    """

    SUCCESS = 0
    UNENCRYPTED_MESSAGE = 1
    DECRYPTION_ERROR = 2
    INCOMPATIBLE_PROTOCOL = 3


class SNDECRYPTConfirm:
    """
    SN-DECRYPT.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.4.3

    ...
    Attributes
    ----------
    plaintext_message_length : int
        Length of the plaintext message (16 bits range)
    plaintext_message : bytes
        Plaintext message
    report : ReportDecrypt
        Decryption report
    """

    def __init__(
        self,
        plaintext_message_length: int,
        plaintext_message: bytes,
        report: ReportDecrypt,
    ):
        self.plaintext_message_length: int = plaintext_message_length
        self.plaintext_message: bytes = plaintext_message
        self.report: ReportDecrypt = report

    def __repr__(self):
        return (
            f"SNDECRYPTConfirm(plaintext_message_length={self.plaintext_message_length}, "
            f"plaintext_message={self.plaintext_message}, report={self.report})"
        )

    def __str__(self):
        return (
            f"SNDECRYPTConfirm(plaintext_message_length={self.plaintext_message_length}, "
            f"plaintext_message={self.plaintext_message}, report={self.report})"
        )


class SNIDCHANGEEVENTCommand(Enum):
    """
    SN-ID-CHANGE-EVENT.command class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.6.2 Table 12 Row 2

    ...
    Attributes
    ----------
    PREPARE
    COMMIT
    ABORT
    DEREG
    """

    PREPARE = 0
    COMMIT = 1
    ABORT = 2
    DEREG = 3


class SNIDCHANGEEVENTIndication:
    """
    SN-ID-CHANGE-EVENT.indication class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.6.2

    ...
    Attributes
    ----------
    command : SNIDCHANGEEVENTCommand
        Command
    identifier : bytes
        ID (8 octets)
    subscriber_data : bytes (optional)
        Subscriber data
    """

    def __init__(
        self,
        command: SNIDCHANGEEVENTCommand,
        identifier: bytes,
        subscriber_data: bytes = None,
    ):
        self.command: SNIDCHANGEEVENTCommand = command
        self.identifier: bytes = identifier
        self.subscriber_data: bytes = subscriber_data

    def __repr__(self):
        return (
            f"SNIDCHANGEEVENTIndication(command={self.command}, identifier={self.identifier}, "
            f"subscriber_data={self.subscriber_data})"
        )

    def __str__(self):
        return (
            f"SNIDCHANGEEVENTIndication(command={self.command}, id={self.identifier}, "
            f"subscriber_data={self.subscriber_data})"
        )


class SNIDCHANGEEVENTResponse:
    """
    SN-ID-CHANGE-EVENT.response class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.6.3

    ...
    Attributes
    ----------
    return_code : bool
        Acknowledgement of the command
    """

    def __init__(self, return_code: bool):
        self.return_code: bool = return_code

    def __repr__(self):
        return f"SNIDCHANGEEVENTResponse(return_code={self.return_code})"

    def __str__(self):
        return f"SNIDCHANGEEVENTResponse(return_code={self.return_code})"


class SNIDCHANGESUBSCRIBERequest:
    """
    SN-ID-CHANGE-SUBSCRIBE.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.5.2

    ...
    Attributes
    ----------
    idchange_event_hook : Callable[[SNIDCHANGEEVENTIndication, bytes], None]
        Callback function to be called when the ID-Change event occurs
    subscriber_data : bytes (optional)
        Subscriber data
    """

    def __init__(
        self,
        idchange_event_hook: Callable[[SNIDCHANGEEVENTIndication, bytes], None],
        subscriber_data: bytes = None,
    ):
        self.idchange_event_hook: Callable[[SNIDCHANGEEVENTIndication, bytes], None] = (
            idchange_event_hook
        )
        self.subscriber_data: bytes = subscriber_data

    def __repr__(self):
        return (
            f"SNIDCHANGESUBSCRIBERequest(idchange_event_hook={self.idchange_event_hook}, "
            f"subscriber_data={self.subscriber_data})"
        )

    def __str__(self):
        return (
            f"SNIDCHANGESUBSCRIBERequest(idchange_event_hook={self.idchange_event_hook}, "
            f"subscriber_data={self.subscriber_data})"
        )


class SNIDCHANGESUBSCRIBEConfirm:
    """
    SN-ID-CHANGE-SUBSCRIBE.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.5.3

    ...
    Attributes
    ----------
    subscription : int
        Subscription handle for unsubscribe (64 bits range)
    """

    def __init__(self, subscription: int):
        self.subscription: int = subscription

    def __repr__(self):
        return f"SNIDCHANGESUBSCRIBEConfirm(subscription={self.subscription})"

    def __str__(self):
        return f"SNIDCHANGESUBSCRIBEConfirm(subscription={self.subscription})"


class SNIDCHANGEUNSUBSCRIBERequest:
    """
    SN-IDCHANGE-UNSUBSCRIBE.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.7.2

    ...
    Attributes
    ----------
    subscription : int
        Subscription handle, given through Mandatory subscribe (64 bits range)
    """

    def __init__(self, subscription: int):
        self.subscription: int = subscription

    def __repr__(self):
        return f"SNIDCHANGEUNSUBSCRIBERequest(subscription={self.subscription})"

    def __str__(self):
        return f"SNIDCHANGEUNSUBSCRIBERequest(subscription={self.subscription})"


class SNIDCHANGEUNSUBSCRIBEConfirm:
    """
    SN-IDCHANGE-UNSUBSCRIBE.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.7.3

    ...
    Empty
    """

    def __repr__(self):
        return "SNIDCHANGEUNSUBSCRIBEConfirm()"

    def __str__(self):
        return "SNIDCHANGEUNSUBSCRIBEConfirm()"


class SNIDCHANGETRIGGERRequest:
    """
    SN-ID-CHANGE-TRIGGER.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.8.2

    ...
    Empty
    """

    def __repr__(self):
        return "SNIDCHANGETRIGGERRequest()"

    def __str__(self):
        return "SNIDCHANGETRIGGERRequest()"


class SNIDCHANGETRIGGERConfirm:
    """
    SN-ID-CHANGE-TRIGGER.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.8.3

    ...
    Empty
    """

    def __repr__(self):
        return "SNIDCHANGETRIGGERConfirm()"

    def __str__(self):
        return "SNIDCHANGETRIGGERConfirm()"


class SNIDLOCKRequest:
    """
    SN-ID-LOCK.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.9.2

    ...
    Attributes
    ----------
    duration : int
        Number of seconds to lock (1 octet)
    """

    def __init__(self, duration: int):
        self.duration: int = duration

    def __repr__(self):
        return f"SNIDLOCKRequest(duration={self.duration})"

    def __str__(self):
        return f"SNIDLOCKRequest(duration={self.duration})"


class SNIDLOCKConfirm:
    """
    SN-ID-LOCK.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.9.3

    ...
    Attributes
    ----------
    lock_handle : int
        Handle to unlock manually (64 bits range)
    """

    def __init__(self, lock_handle: int):
        self.lock_handle: int = lock_handle

    def __repr__(self):
        return f"SNIDLOCKConfirm(lock_handle={self.lock_handle})"

    def __str__(self):
        return f"SNIDLOCKConfirm(lock_handle={self.lock_handle})"


class SNIDUNLOCKRequest:
    """
    SN-ID-UNLOCK.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.10.2

    ...
    Attributes
    ----------
    lock_handle : int
        Handle to unlock manually (64 bits range)
    """

    def __init__(self, lock_handle: int):
        self.lock_handle: int = lock_handle

    def __repr__(self):
        return f"SNIDUNLOCKRequest(lock_handle={self.lock_handle})"

    def __str__(self):
        return f"SNIDUNLOCKRequest(lock_handle={self.lock_handle})"


class SNIDUNLOCKConfirm:
    """
    SN-ID-UNLOCK.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.10.3

    ...
    Empty
    """

    def __repr__(self):
        return "SNIDUNLOCKConfirm()"

    def __str__(self):
        return "SNIDUNLOCKConfirm()"


class SNLOGSECURITYEVENTEventType(Enum):
    """
    SN-LOG-SECURITY-EVENT.event-type as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.11.2

    ...
    Attributes
    ----------
    TIME_CONSISTENCY_FAILED
    LOCATION_CONSISTENCY_FAILED
    ID_CONSISTENCY_FAILED
    DISALLOWED_MESSAGE_CONTENT
    DISALLOWED_MESSAGE_FREQUENCY
    REPLAY_DETECTION_TIME
    REPLAY_DETECTION_LOCATION
    MOVEMENT_PLAUSIBILITY
    APPEARANCE_PLAUSIBILITY
    LOCATION_PLAUSIBILITY_SENSOR
    LOCATION_PLAUSIBILITY_MAP
    LOCATION_PLAUSIBILITY_CONTRADICTION
    LOCATION_PLAUSIBILITY_CONTRADICTION_VEHICLE_DIMENSION
    LOCATION_PLAUSIBILITY_CONTRADICTION_NEIGHBOR_INFO
    """

    TIME_CONSISTENCY_FAILED = 0x01
    LOCATION_CONSISTENCY_FAILED = 0x02
    ID_CONSISTENCY_FAILED = 0x03
    DISALLOWED_MESSAGE_CONTENT = 0x04
    DISALLOWED_MESSAGE_FREQUENCY = 0x05
    REPLAY_DETECTION_TIME = 0x06
    REPLAY_DETECTION_LOCATION = 0x07
    MOVEMENT_PLAUSIBILITY = 0x08
    APPEARANCE_PLAUSIBILITY = 0x09
    LOCATION_PLAUSIBILITY_SENSOR = 0x0A
    LOCATION_PLAUSIBILITY_MAP = 0x0B
    LOCATION_PLAUSIBILITY_CONTRADICTION = 0x0C
    LOCATION_PLAUSIBILITY_CONTRADICTION_VEHICLE_DIMENSION = 0x0D
    LOCATION_PLAUSIBILITY_CONTRADICTION_NEIGHBOR_INFO = 0x0E


class SNLOGSECURITYEVENTEventEvidenceType(Enum):
    """
    SN-LOG-SECURITY-EVENT.event-evidence-type as specified in ETSI TS 102 723-8 V1.1.1 (2016-04)

    ...
    Attributes
    ----------
    CAM
    DENM
    """

    CAM = 0x01
    DENM = 0x02


class SNLOGSECURITYEVENTRequest:
    # pylint: disable=too-many-arguments
    """
    SN-LOG-SECURITY-EVENT.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.11.2

    ...
    Attributes
    ----------
    event_type : SNLOGSECURITYEVENTEventType
        Type of security event
    neighbour_id_list_length : int
        Length of neighbour ID list (1 octet)
    neighbour_id_list : list[bytes]
        List of affected V2X neighbour nodes, expressed via certificate hash
    event_time : int
        Timestamp of security event (4 octets)
    event_location : dict (optional)
        Location of security event in format {"latitude": float, "longitude": float}
    event_evidence_list_length : int (optional)
        Length of event evidence list (4 octets)
    event_evidence_list : list[dict] (optional)
        List of event evidence in format {"length": int, "data": bytes}
    event_evidence_type : SNLOGSECURITYEVENTEventEvidenceType (optional)
        Type of event evidence
    event_evidence_content_length : int (optional)
        Length of event evidence content (4 octets)
    event_evidence_content : bytes (optional)
        Event evidence content
    """

    def __init__(
        self,
        event_type: SNLOGSECURITYEVENTEventType,
        neighbour_id_list_length: int,
        neighbour_id_list: list[bytes],
        event_time: int,
        event_location: dict = None,
        event_evidence_list_length: int = None,
        event_evidence_list: list[dict] = None,
        event_evidence_type: SNLOGSECURITYEVENTEventEvidenceType = None,
        event_evidence_content_length: int = None,
        event_evidence_content: bytes = None,
    ):
        self.event_type: SNLOGSECURITYEVENTEventType = event_type
        self.neighbour_id_list_length: int = neighbour_id_list_length
        self.neighbour_id_list: list[bytes] = neighbour_id_list
        self.event_time: int = event_time
        self.event_location: dict = event_location
        self.event_evidence_list_length: int = event_evidence_list_length
        self.event_evidence_list: list[dict] = event_evidence_list
        self.event_evidence_type: SNLOGSECURITYEVENTEventEvidenceType = (
            event_evidence_type
        )
        self.event_evidence_content_length: int = event_evidence_content_length
        self.event_evidence_content: bytes = event_evidence_content

    def __repr__(self):
        return (
            f"SNLOGSECURITYEVENTRequest(event_type={self.event_type}, "
            f"neighbour_id_list_length={self.neighbour_id_list_length}, "
            f"neighbour_id_list={self.neighbour_id_list}, event_time={self.event_time}, "
            f"event_location={self.event_location}, event_evidence_list_length={self.event_evidence_list_length}, "
            f"event_evidence_list={self.event_evidence_list}, event_evidence_type={self.event_evidence_type}, "
            f"event_evidence_content_length={self.event_evidence_content_length}, "
            f"event_evidence_content={self.event_evidence_content})"
        )

    def __str__(self):
        return (
            f"SNLOGSECURITYEVENTRequest(event_type={self.event_type}, "
            f"neighbour_id_list_length={self.neighbour_id_list_length}, "
            f"neighbour_id_list={self.neighbour_id_list}, event_time={self.event_time}, "
            f"event_location={self.event_location}, event_evidence_list_length={self.event_evidence_list_length}, "
            f"event_evidence_list={self.event_evidence_list}, event_evidence_type={self.event_evidence_type}, "
            f"event_evidence_content_length={self.event_evidence_content_length}, "
            f"event_evidence_content={self.event_evidence_content})"
        )


class SNLOGSECURITYEVENTConfirm:
    """
    SN-LOG-SECURITY-EVENT.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.11.3

    ...
    Empty
    """

    def __init__(self):
        pass

    def __repr__(self):
        return "SNLOGSECURITYEVENTConfirm()"

    def __str__(self):
        return "SNLOGSECURITYEVENTConfirm()"


class SNENCAPRequest:
    """
    SN-ENCAP.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.12.2

    ...
    Attributes
    ----------
    tbe_packet_length : int
        Length of TBE packet (2 octets)
    tbe_packet : bytes
        TBE packet
    sec_services : int (optional)
        Security services to invoke (2 octets)
    its_aid_length : int (optional)
        Length of ITS AID (2 octets)
    its_aid : int
        ITS AID
    permissions : bytes
        SSP associated with the ITS AID (Max length 31 octets)
    context_information : bytes (optional)
        Context information
    target_id_list_length : int (optional)
        Length of target ID list (2 octets)
    target_id_list : list[bytes] (optional)
        List of target IDs
    """

    def __init__(
        self,
        tbe_packet_length: int,
        tbe_packet: bytes,
        sec_services: int = None,
        its_aid_length: int = None,
        its_aid: int = None,
        permissions: bytes = None,
        context_information: bytes = None,
        target_id_list_length: int = None,
        target_id_list: list[bytes] = None,
    ):
        self.tbe_packet_length: int = tbe_packet_length
        self.tbe_packet: bytes = tbe_packet
        self.sec_services: int = sec_services
        self.its_aid_length: int = its_aid_length
        self.its_aid: int = its_aid
        self.permissions: bytes = permissions
        self.context_information: bytes = context_information
        self.target_id_list_length: int = target_id_list_length
        self.target_id_list: list[bytes] = target_id_list

    def __repr__(self):
        return (
            f"SNENCAPRequest(tbe_packet_length={self.tbe_packet_length}, tbe_packet={self.tbe_packet}, "
            f"sec_services={self.sec_services}, its_aid_length={self.its_aid_length}, its_aid={self.its_aid}, "
            f"permissions={self.permissions}, context_information={self.context_information}, "
            f"target_id_list_length={self.target_id_list_length}, target_id_list={self.target_id_list})"
        )

    def __str__(self):
        return (
            f"SNENCAPRequest(tbe_packet_length={self.tbe_packet_length}, tbe_packet={self.tbe_packet}, "
            f"sec_services={self.sec_services}, its_aid_length={self.its_aid_length}, its_aid={self.its_aid}, "
            f"permissions={self.permissions}, context_information={self.context_information}, "
            f"target_id_list_length={self.target_id_list_length}, target_id_list={self.target_id_list})"
        )


class SNENCAPConfirm:
    """
    SN-ENCAP.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.12.3

    ...
    Attributes
    ----------
    sec_packet_length : int
        Length of security packet (2 octets)
    sec_packet : bytes
        Security packet
    """

    def __init__(self, sec_packet_length: int, sec_packet: bytes):
        self.sec_packet_length: int = sec_packet_length
        self.sec_packet: bytes = sec_packet

    def __repr__(self):
        return f"SNENCAPConfirm(sec_packet_length={self.sec_packet_length}, sec_packet={self.sec_packet})"

    def __str__(self):
        return f"SNENCAPConfirm(sec_packet_length={self.sec_packet_length}, sec_packet={self.sec_packet})"


class SNDECAPRequest:
    """
    SN-DECAP.request class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.13.2

    ...
    Attributes
    ----------
    sec_packet_length : int
        Length of security packet (2 octets)
    sec_packet : bytes
        Security packet
    """

    def __init__(self, sec_packet_length: int, sec_packet: bytes):
        self.sec_packet_length: int = sec_packet_length
        self.sec_packet: bytes = sec_packet

    def __repr__(self):
        return f"SNDECAPRequest(sec_packet_length={self.sec_packet_length}, sec_packet={self.sec_packet})"

    def __str__(self):
        return f"SNDECAPRequest(sec_packet_length={self.sec_packet_length}, sec_packet={self.sec_packet})"


class SNDECAPReport(Enum):
    """
    SN-DECAP.report class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.13.3 Table 27 row 4

    ...
    Attributes
    ----------
    SUCCESS
    FALSE_SIGNATURE
    INVALID_CERTIFICATE
    REVOKED_CERTIFICATE
    INCONSISTENT_CHAIN
    INVALID_TIMESTAMP
    DUPLICATE_MESSAGE
    INVALID_MOBILITY_DATA
    UNSIGNED_MESSAGE
    SIGNER_CERTIFICATE_NOT_FOUND
    UNSUPPORTED_SIGNER_IDENTIFIER_TYPE
    INCOMPATIBLE_PROTOCOL
    UNENCRYPTED_MESSAGE
    DECRYPTION_ERROR
    """

    SUCCESS = 0
    FALSE_SIGNATURE = 1
    INVALID_CERTIFICATE = 2
    REVOKED_CERTIFICATE = 3
    INCONSISTENT_CHAIN = 4
    INVALID_TIMESTAMP = 5
    DUPLICATE_MESSAGE = 6
    INVALID_MOBILITY_DATA = 7
    UNSIGNED_MESSAGE = 8
    SIGNER_CERTIFICATE_NOT_FOUND = 9
    UNSUPPORTED_SIGNER_IDENTIFIER_TYPE = 10
    INCOMPATIBLE_PROTOCOL = 11
    UNENCRYPTED_MESSAGE = 12
    DECRYPTION_ERROR = 13


class SNDECAPConfirm:
    """
    SN-DECAP.confirm class as specified in ETSI TS 102 723-8 V1.1.1 (2016-04) 5.2.13.3

    ...
    Attributes
    ----------
    plaintext_packet_length : int
        Length of plaintext packet (2 octets)
    plaintext_packet : bytes
        Plaintext packet
    report: SNDECAPReport
        Report
    certificate_id : bytes (optional)
        Certificate ID (8 octets)
    its_aid_length : int
        Length of ITS AID (2 octets)
    its_aid : int
        ITS AID
    permissions : bytes (optional)
        SSP associated with the ITS AID (Max length 31 octets)
    """

    def __init__(
        self,
        plaintext_packet_length: int,
        plaintext_packet: bytes,
        report: SNDECAPReport,
        certificate_id: bytes = None,
        its_aid_length: int = None,
        its_aid: int = None,
        permissions: bytes = None,
    ):
        self.plaintext_packet_length: int = plaintext_packet_length
        self.plaintext_packet: bytes = plaintext_packet
        self.report: SNDECAPReport = report
        self.certificate_id: bytes = certificate_id
        self.its_aid_length: int = its_aid_length
        self.its_aid: int = its_aid
        self.permissions: bytes = permissions

    def __repr__(self):
        return (
            f"SNDECAPConfirm(plaintext_packet_length={self.plaintext_packet_length}, "
            f"plaintext_packet={self.plaintext_packet}, report={self.report}, "
            f"certificate_id={self.certificate_id}, its_aid_length={self.its_aid_length}, "
            f"its_aid={self.its_aid}, permissions={self.permissions})"
        )

    def __str__(self):
        return (
            f"SNDECAPConfirm(plaintext_packet_length={self.plaintext_packet_length}, "
            f"plaintext_packet={self.plaintext_packet}, report={self.report}, "
            f"certificate_id={self.certificate_id}, its_aid_length={self.its_aid_length}, "
            f"its_aid={self.its_aid}, permissions={self.permissions})"
        )



from __future__ import annotations
import threading

from .ldm_maintenance import LDMMaintenance
from .ldm_service import LDMService


class LDMServiceThreads(LDMService):
    """
    Class that inherits from LDMService (class specified in the ETSI ETSI EN 302 895 V1.1.1 (2014-09).

    Class that is used to run the service of the LDM in a seperate thread. This is done by running the
    subscription_service function in a seperate thread.
    """

    def __init__(
        self,
        ldm_maintenance: LDMMaintenance,
        stop_event: threading.Event = None,
    ) -> None:
        super().__init__(ldm_maintenance)
        self.data_containers_lock = threading.Lock()
        self.stop_event = stop_event
        if stop_event is None:
            self.stop_event = threading.Event()

        self.subscriptions_service_thread = threading.Thread(
            target=super().subscriptions_service, daemon=True
        )
        self.subscriptions_service_thread.start()

    def subscriptions_service(self) -> None:
        while not self.stop_event.is_set():
            self.attend_subscriptions()



from base64 import b64encode, b64decode
from ..geonet.gn_address import GNAddress
from ..geonet.service_access_point import (
    Area,
    GNDataIndication,
    PacketTransportType,
    CommunicationProfile,
    TrafficClass,
    CommonNH,
)
from ..geonet.position_vector import LongPositionVector


class BTPDataRequest:
    """
    GN Data Request class. As specified in
    ETSI EN 302 636-5-1 V2.1.0 (2017-05). Annex A2

    Attributes
    ----------
    btp_type : CommonNH
        BTP Type.
    source_port : int
        (16 bit integer) Source Port.
    destination_port : int
        (16 bit integer) Destination Port.
    destinaion_port_info : int
        (16 bit integer) Destination Port Info.
    gn_packet_transport_type : PacketTransportType
        Packet Transport Type.
    gn_destination_address : GNAddress
        Destination Address.
    communication_profile : CommunicationProfile
        Communication Profile.
    traffic_class : TrafficClass
        Traffic Class.
    length : int
        Length of the payload.
    data : bytes
        Payload.
    """

    def __init__(self) -> None:
        self.btp_type = CommonNH.BTP_B
        self.source_port = 0
        self.destination_port = 0
        self.destinaion_port_info = 0
        self.gn_packet_transport_type = PacketTransportType()
        self.gn_destination_address = GNAddress()
        self.gn_area = Area()
        self.communication_profile = CommunicationProfile.UNSPECIFIED
        self.traffic_class = TrafficClass()
        self.length = 0
        self.data = b""

    def to_dict(self) -> dict:
        """
        Returns the BTPDataRequest as a dictionary.

        Returns
        -------
        dict
            Dictionary representation of the BTPDataRequest.
        """
        return {
            "btp_type": self.btp_type.value,
            "source_port": self.source_port,
            "destination_port": self.destination_port,
            "destinaion_port_info": self.destinaion_port_info,
            "gn_packet_transport_type": self.gn_packet_transport_type.to_dict(),
            "gn_destination_address": b64encode(
                self.gn_destination_address.encode()
            ).decode("utf-8"),
            "gn_area": self.gn_area.to_dict(),
            "communication_profile": self.communication_profile.value,
            "traffic_class": b64encode(self.traffic_class.encode_to_bytes()).decode(
                "utf-8"
            ),
            "length": self.length,
            "data": b64encode(self.data).decode("utf-8"),
        }

    def from_dict(self, data: dict) -> None:
        """
        Initializes the BTPDataRequest with a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary to initialize with.
        """
        self.btp_type = CommonNH(data["btp_type"])
        self.source_port = data["source_port"]
        self.destination_port = data["destination_port"]
        self.destinaion_port_info = data["destinaion_port_info"]
        self.gn_packet_transport_type = PacketTransportType()
        self.gn_packet_transport_type.from_dict(data["gn_packet_transport_type"])
        self.gn_destination_address = GNAddress()
        self.gn_destination_address.decode(b64decode(data["gn_destination_address"]))
        self.gn_area = Area()
        self.gn_area.from_dict(data["gn_area"])
        self.communication_profile = CommunicationProfile(data["communication_profile"])
        self.traffic_class = TrafficClass()
        self.traffic_class.decode_from_bytes(b64decode(data["traffic_class"]))
        self.length = data["length"]
        self.data = b64decode(data["data"])


class BTPDataIndication:
    """
    GN Data Indication class. As specified in ETSI EN 302 636-5-1 V2.1.0 (2017-05). Annex A3

    Attributes
    ----------
    source_port : int
        (16 bit integer) Source Port.
    destination_port : int
        (16 bit integer) Destination Port.
    destinaion_port_info : int
        (16 bit integer) Destination Port Info.
    gn_packet_transport_type : PacketTransportType
        Packet Transport Type.
    gn_destination_address : GNAddress
        Destination Address.
    gn_source_position_vector : LongPositionVector
        Source Position Vector.
    gn_traffic_class : TrafficClass
        Traffic Class.
    length : int
        Length of the payload.
    data : bytes
        Payload.
    """

    def __init__(self) -> None:
        self.source_port = 0
        self.destination_port = 0
        self.destinaion_port_info = 0
        self.gn_packet_transport_type = PacketTransportType()
        self.gn_destination_address = GNAddress()
        self.gn_source_position_vector = LongPositionVector()
        self.gn_traffic_class = TrafficClass()
        self.length = 0
        self.data = b""

    def initialize_with_gn_data_indication(
        self, gn_data_indication: GNDataIndication
    ) -> None:
        """
        Initializes the BTPDataIndication with a GNDataIndication.

        Parameters
        ----------
        gn_data_indication : GNDataIndication
            GNDataIndication to initialize with.
        """
        self.gn_packet_transport_type = gn_data_indication.packet_transport_type
        self.gn_source_position_vector = gn_data_indication.source_position_vector
        self.gn_traffic_class = gn_data_indication.traffic_class
        self.data = gn_data_indication.data[4:]
        self.length = len(self.data)

    def to_dict(self) -> dict:
        """
        Returns the BTPDataIndication as a dictionary.

        Returns
        -------
        dict
            Dictionary representation of the BTPDataIndication.
        """
        return {
            "source_port": self.source_port,
            "destination_port": self.destination_port,
            "destinaion_port_info": self.destinaion_port_info,
            "gn_packet_transport_type": self.gn_packet_transport_type.to_dict(),
            "gn_destination_address": b64encode(
                self.gn_destination_address.encode()
            ).decode("utf-8"),
            "gn_source_position_vector": b64encode(
                self.gn_source_position_vector.encode()
            ).decode("utf-8"),
            "gn_traffic_class": b64encode(
                self.gn_traffic_class.encode_to_bytes()
            ).decode("utf-8"),
            "length": self.length,
            "data": b64encode(self.data).decode("utf-8"),
        }

    def from_dict(self, data: dict) -> None:
        """
        Initializes the BTPDataIndication with a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary to initialize with.
        """
        self.source_port = data["source_port"]
        self.destination_port = data["destination_port"]
        self.destinaion_port_info = data["destinaion_port_info"]
        self.gn_packet_transport_type = PacketTransportType()
        self.gn_packet_transport_type.from_dict(data["gn_packet_transport_type"])
        self.gn_destination_address = GNAddress()
        self.gn_destination_address.decode(b64decode(data["gn_destination_address"]))
        self.gn_source_position_vector = LongPositionVector()
        self.gn_source_position_vector.decode(
            b64decode(data["gn_source_position_vector"])
        )
        self.gn_traffic_class = TrafficClass()
        self.gn_traffic_class.decode_from_bytes(b64decode(data["gn_traffic_class"]))
        self.length = data["length"]
        self.data = b64decode(data["data"])



from __future__ import annotations
from collections.abc import Callable
import logging
from .ldm_constants import (
    DENM,
    DATA_OBJECT_TYPE_ID,
    MAPEM,
    SPATEM,
    VALID_ITS_AID,
)
from .ldm_classes import (
    Filter,
    OrderTuple,
    RegisterDataConsumerReq,
    RequestDataObjectsReq,
    RequestDataObjectsResp,
    SubscribeDataobjectsReq,
    SubscribeDataObjectsResp,
    DeregisterDataConsumerAck,
    DeregisterDataConsumerReq,
    SubscribeDataobjectsResult,
    RegisterDataConsumerResp,
    RequestedDataObjectsResult,
    DeregisterDataConsumerResp,
    RegisterDataConsumerResult,
    TimestampIts,
    UnsubscribeDataConsumerReq,
    UnsubscribeDataConsumerResp,
    UnsubscribeDataConsumerAck,
)
from .ldm_service import LDMService


class InterfaceLDM4:
    """
    Class specified is described in ETSI EN 302 895 V1.1.1 (2014-09). Section 5.4.4.
    The LDM shall provide an interface IF.LDM.4 to enable an application or facility to register as a LDM Data Consumer

    Data consumers, in our implementation, ask for data in the form of dataframes,
    as specified in ETSI TS 102 894-2 V1.3.1 (2018-08) and updated in https://forge.etsi.org/rep/ITS/asn1.

    Attributes
    ----------
    TODO: Add attributes
    TODO: There is replicated code (IF_LDM_3 and IF_LDM_4). Check if it can be removed.

    """

    def __init__(self, ldm_service: LDMService) -> None:
        self.logging = logging.getLogger("local_dynamic_map")

        self.ldm_service = ldm_service

    def check_its_aid(self, its_application_identifier) -> bool:
        """
        Method specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.2.1
        It checks the ITS-AID of the application that wants to register as a data provider.

        Parameters
        ----------
        its_application_identifier : int
            ITS-AID of the application that wants to register as a data provider
        """
        return (
            isinstance(its_application_identifier, int)
            and its_application_identifier in VALID_ITS_AID
        )

    def check_permissions(self, permissions_granted: list, data_object_id: int) -> bool:
        """
        Method that checks permissions to grant access to the data provider
        as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.2.1.
        TODO: Implement better security. Currenlty very basic security is implemented.

        Parameters
        ----------
        permissions_granted : list
            List of permissions granted to the data provider
        dataObjectID : int
        """
        if isinstance(permissions_granted, list):
            if len(permissions_granted) == 0:
                return False
            for permission in permissions_granted:
                if permission == data_object_id:
                    return True
                # TODO: Change SPATEM and MAPEM!!! DENMs have access to all data objects.
                if data_object_id in (DENM, SPATEM, MAPEM):
                    return True
        return False

    def register_data_consumer(
        self, data_consumer: RegisterDataConsumerReq
    ) -> RegisterDataConsumerResp:
        """
        Method specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.1.1
        TODO: Apply Security. Currenly gives access to anyone sending valid information.
        TODO: Validate check permissions and its_aid.

        Parameters
        ----------
        data_consumer : RegisterDataConsumerReq
            RegisterDataConsumerReq as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section Annex B.
        """
        if (
            data_consumer is None
            or self.check_its_aid(data_consumer.application_id) is False
            or self.check_permissions(
                data_consumer.access_permisions, data_consumer.application_id
            )
            is False
        ):
            return RegisterDataConsumerResp(
                data_consumer.application_id, None, RegisterDataConsumerResult(2)
            )
        self.ldm_service.add_data_consumer_its_aid(data_consumer.application_id)
        self.logging.debug(
            "Registered new LDM Data Consumer with application_id %d",
            data_consumer.application_id,
        )
        return RegisterDataConsumerResp(
            data_consumer.application_id,
            data_consumer.access_permisions,
            RegisterDataConsumerResult(0),
        )

    def deregister_data_consumer(
        self, data_consumer: DeregisterDataConsumerReq
    ) -> DeregisterDataConsumerResp:
        """
        Method specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.2.1
        Returning ACK 0 if the application is registered (and now will be unregistered) and
        ACK 1 if it is not (and will not be unregistered).
        """
        self.logging.debug(
            "Deregistring LDM Data Consumer wtih application_id: %d",
            data_consumer.application_id,
        )

        if data_consumer.application_id in self.ldm_service.get_data_consumer_its_aid():
            self.ldm_service.del_data_consumer_its_aid(data_consumer.application_id)
            return DeregisterDataConsumerResp(
                data_consumer.application_id, DeregisterDataConsumerAck(0)
            )
        return DeregisterDataConsumerResp(
            data_consumer.application_id, DeregisterDataConsumerAck(1)
        )

    def request_data_objects(
        self, data_request: RequestDataObjectsReq
    ) -> RequestDataObjectsResp:
        """
        Method specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.3
        TODO: Add error message in response if result is not successful.
        TODO: Raise errors! Create an expection file!
        Parameters
        ----------
        data_request : RequestDataObjectsReq
            RequestDataObjectsReq as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section Annex B.
        """
        self.logging.debug(
            "LDM Data Consumer with application id %s, asking for the following data types; %s",
            data_request.application_id,
            data_request.data_object_type,
        )
        if (
            data_request.application_id
            not in self.ldm_service.get_data_consumer_its_aid()
        ):
            return RequestDataObjectsResp(
                data_request.application_id, None, RequestedDataObjectsResult(1)
            )
        for data_object_type in data_request.data_object_type:
            if data_object_type not in DATA_OBJECT_TYPE_ID:
                return RequestDataObjectsResp(
                    data_request.application_id, None, RequestedDataObjectsResult(2)
                )

        if data_request.priority is not None:
            if data_request.priority < 0 or data_request.priority > 255:
                return RequestDataObjectsResp(
                    data_request.application_id, None, RequestedDataObjectsResult(3)
                )
        if data_request.order is not None:
            if not isinstance(data_request.order, list):
                return RequestDataObjectsResp(
                    data_request.application_id,
                    None,
                    RequestedDataObjectsResult(5),
                )
        if data_request.filter is not None:
            if not isinstance(data_request.filter, Filter):
                return RequestDataObjectsResp(
                    data_request.application_id, None, RequestedDataObjectsResult(4)
                )
        data_objects: list = self.ldm_service.query(
            data_request
        )  # Find data objects in LDM Maintance DataBases
        self.logging.debug(
            "LDM Data Consumer with application id %s, has recieved %s.",
            data_request.application_id,
            len(data_objects),
        )
        return RequestDataObjectsResp(
            data_request.application_id, data_objects, RequestedDataObjectsResult(0)
        )

    def subscribe_data_consumer(
        self,
        subscribe_data_consumer: SubscribeDataobjectsReq,
        callback: Callable[[RequestDataObjectsResp], None],
    ) -> SubscribeDataObjectsResp:
        """
        Method specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4.1
        Method parameters have been modified in order to implement standard.
        Currently, the "callback" parameter is used to "PublishDataObject".
        New Version -> SubscribeDataobjectsReq(applicationId,
                                                dataObjects,
                                                callback,
                                                priority,
                                                order,
                                                filter,
                                                notifyTime,
                                                multiplicity,
                                                order)
        Standard Version -> SubscribeDataobjectsReq(applicationId,
                                                    dataObjects,
                                                    priority,
                                                    order,
                                                    filter,
                                                    notifyTime,
                                                    multiplicity,
                                                    order)

        Parameters
        ----------
        subscribe_data_consumer : SubscribeDataobjectsReq
            SubscribeDataobjectsReq as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section Annex B.
        callback : Callable[[None], None]
            Callback function that will be called when data is available for the data consumer.
        """
        self.logging.debug(
            "LDM Data Consumer subscribed with application id %s",
            str(subscribe_data_consumer),
        )
        result = self.validate_subscribe_data_consumer(subscribe_data_consumer)

        if result is not None:
            return result

        subscription_id = self.store_subscription_info(
            subscribe_data_consumer, callback
        )

        return SubscribeDataObjectsResp(
            subscribe_data_consumer.application_id,
            subscription_id,
            SubscribeDataobjectsResult(0),
            None,
        )

    def validate_subscribe_data_consumer(
        self,
        subscribe_data_consumer: SubscribeDataobjectsReq,
    ) -> SubscribeDataObjectsResp:
        """
        Method that groups all the validation methods for the SubscribeDataobjectsReq

        Parameters
        ----------
        subscribe_data_consumer : SubscribeDataobjectsReq
            SubscribeDataobjectsReq as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section Annex B.

        Returns
        -------
        Optional[SubscribeDataObjectsResp]
            SubscribeDataObjectsResp with error message if there is an error, None otherwise.
        """
        result_code = None
        error_message = None

        if not self.is_valid_its_aid(subscribe_data_consumer.application_id):
            result_code = 1
            error_message = "Invalid ITS-AID"

        if not self.is_valid_data_object_type(subscribe_data_consumer.data_object_type):
            result_code = 2
            error_message = "Invalid data object type"

        if not self.is_valid_priority(subscribe_data_consumer.priority):
            result_code = 3
            error_message = "Invalid priority"

        if not self.is_valid_order(subscribe_data_consumer.order):
            result_code = 7
            error_message = "Invalid order"

        if not self.is_valid_filter(subscribe_data_consumer.filter):
            result_code = 4
            error_message = "Invalid filter"

        if not self.is_valid_notify_time(subscribe_data_consumer.notify_time):
            result_code = 5
            error_message = "Invalid notifyTime"

        if not self.is_valid_multiplicity(subscribe_data_consumer.multiplicity):
            result_code = 6
            error_message = "Invalid multiplicity"

        if result_code is not None:
            return self.invalid_response(
                subscribe_data_consumer.application_id, result_code, error_message
            )

        return None

    def invalid_response(
        self, application_id: int, result_code: int, error_message: str
    ) -> SubscribeDataObjectsResp:
        """
        Method to create SubscribeDataObjectResp with error message as specified
        in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        application_id : int
            Application ID of the data consumer
        result_code : int
            Result code of the response
        error_message : str
            Error message to be included in the response

        Returns
        -------
        SubscribeDataObjectsResp
            SubscribeDataObjectsResp with error message
        """
        return SubscribeDataObjectsResp(
            application_id,
            None,
            SubscribeDataobjectsResult(result_code),
            error_message,
        )

    def is_valid_its_aid(self, application_id: int) -> bool:
        """
        Method to check if application_id (its aid) is valid as specified
        in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        application_id : int
            Application ID to be checked

        Returns
        -------
        bool
            True if application_id is valid, False otherwise.
        """
        return application_id in self.ldm_service.get_data_consumer_its_aid()

    def is_valid_data_object_type(self, data_object_type: str) -> bool:
        """
        Method to check if data_object_types are valid as specified in CDD TS 102 894-2.

        Parameters
        ----------
        data_object_types : list
            List of data object types to be checked

        Returns
        -------
        bool
            True if data_object_types are valid, False otherwise.
        """
        return all(
            data_object_type in DATA_OBJECT_TYPE_ID
            for data_object_type in data_object_type
        )

    def is_valid_priority(self, priority: int) -> bool:
        """
        Method to check if priority is valid as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        priority : int
            Priority to be checked

        Returns
        -------
        bool
            True if priority is valid, False otherwise.
        """
        return priority is None or (0 <= priority <= 255)

    def is_valid_order(self, order: list[OrderTuple]) -> bool:
        """
        Method to check if order is valid as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        order : int
            Order to be checked

        Returns
        -------
        bool
            True if order is valid, False otherwise.
        """
        return order is None or (
            isinstance(order, list) and all(o in [1, 2] for o in order)
        )

    def is_valid_filter(self, data_filter: Filter) -> bool:
        """
        Method to check if filter is valid as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        data_filter : Filter
            Filter to be checked

        Returns
        -------
        bool
            True if filter is valid, False otherwise.
        """
        return data_filter is None or isinstance(data_filter, Filter)

    def is_valid_notify_time(self, notify_time: TimestampIts) -> bool:
        """
        Method to check if notifyTime is valid as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        notify_time : int
            NotifyTime to be checked

        Returns
        -------
        bool
            True if notifyTime is valid, False otherwise.
        """
        return notify_time is None or (0 <= notify_time <= 4398046511103)

    def is_valid_multiplicity(self, multiplicity: int) -> bool:
        """
        Method to check if the multiplicity value is valid as specified
        in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        multiplicity : int
            Multiplicity to be checked

        Returns
        -------
        bool
            True if multiplicity is valid, False otherwise.
        """
        return multiplicity is None or (0 <= multiplicity <= 255)

    def store_subscription_info(
        self, subscribe_data_consumer: SubscribeDataobjectsReq, callback: Callable
    ) -> int:
        """
        Method to store subscription information as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.4

        Parameters
        ----------
        subscribe_data_consumer : SubscribeDataobjectsReq
            SubscribeDataobjectsReq as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section Annex B.
        callback : Callable[[None], None]
            Callback function that will be called when data is available for the data consumer.

        Returns
        -------
        int
            Subscription ID (index of the subscription in the database)
        """
        return self.ldm_service.store_new_subscription_petition(
            subscribe_data_consumer.application_id,
            subscribe_data_consumer.data_object_type,
            subscribe_data_consumer.priority,
            subscribe_data_consumer.filter,
            subscribe_data_consumer.notify_time,
            subscribe_data_consumer.multiplicity,
            subscribe_data_consumer.order,
            callback,
        )

    def unsubscribe_data_consumer(
        self, unsubscribe_data_consumer: UnsubscribeDataConsumerReq
    ):
        """
        Method specified in ETSI EN 302 895 V1.1.1 (2014-09). Section 6.3.5.1
        Current standard doesn't have ASN.1 implementation for UnsubscribeDataConsumerReq
        or UnsubscribeDataConsumerResp so it has been implemented in-house.
        Will have to be updated once standard is updated.

        Parameters
        ----------
        unsubscribe_data_consumer : UnsubscribeDataConsumerReq
            UnsubscribeDataConsumerReq as specified in ETSI EN 302 895 V1.1.1 (2014-09). Section Annex B.
        """
        self.logging.debug(
            "LDM Data Consumer Subscriber with application_id %s has unsubscribed.",
            unsubscribe_data_consumer.application_id,
        )
        if (
            unsubscribe_data_consumer.application_id
            not in self.ldm_service.get_data_consumer_its_aid()
        ):
            return UnsubscribeDataConsumerResp(
                unsubscribe_data_consumer.application_id,
                None,
                UnsubscribeDataConsumerAck(1),
            )

        if not isinstance(unsubscribe_data_consumer.subscription_id, int):
            return UnsubscribeDataConsumerResp(
                unsubscribe_data_consumer.application_id,
                unsubscribe_data_consumer.subscription_id,
                UnsubscribeDataConsumerAck(1),
            )

        if not self.ldm_service.delete_subscription(
            unsubscribe_data_consumer.subscription_id
        ):
            return UnsubscribeDataConsumerResp(
                unsubscribe_data_consumer.application_id,
                unsubscribe_data_consumer.subscription_id,
                UnsubscribeDataConsumerAck(1),
            )

        return UnsubscribeDataConsumerResp(
            unsubscribe_data_consumer.application_id,
            unsubscribe_data_consumer.subscription_id,
            UnsubscribeDataConsumerAck(0),
        )



from __future__ import annotations
from collections.abc import Callable


class LinkLayer:
    """
    Link Layer class to connect layer 2 and layer 3.

    Attributes
    ----------
    receive_callback : Callable[[bytes], None]
        Callback function to receive packets.

    Methods
    -------
    send(bytes)
        Send a packet to the LL.
    receive()
        Receive a packet from the LL. (To be called in a thread)


    """

    def __init__(self, receive_callback: Callable[[bytes], None]) -> None:
        """
        Create a Link Layer object.

        Parameters
        ----------
        iface : str
            Interface to use. (Linux only)
        mac_address : bytes
            MAC address of the interface.
        receive_callback : Callable[[bytes], None]
            Callback function to receive packets.

        Raises
        ------
        OSError:
            If the interface is not found.
        Exception:
            If the MAC address is not 6 bytes long.
        """
        self.receive_callback = receive_callback

    def send(self, packet: bytes) -> None:
        """
        Send a packet to the lower layer.

        Parameters
        ----------
        packet : bytes
            Packet to send.

        Raises
        ------
        SendingException:
            If the packet cannot be sent.
        PacketTooLongException:
            If the packet is too long.
        NotImplementedError:
            If the method is not implemented.
        """
        raise NotImplementedError("All the Link Layers should implement this")



from .ldm_classes import RequestDataObjectsReq


class DataBase:
    """
    Generic DataBase class that will be implemented by various classes (i.e. TinyDBDatabase, ListDatabase).
    """
    def __init__(self, database_name: str = None, database_path: str = None):
        """
        Create database with a given name, returns a boolean stating if creation has been succesful.

        Parameters
        ----------
        database_name : str
            Name of the database to be created.
        """

    def delete(self, database_name: str = None) -> bool:
        """
        Delete database with a given name, returns a boolean stating if deletion has been succesful.

        Parameters
        ----------
        database_name : str
            Name of the database to be deleted.
        """

    def search(self, data_request: RequestDataObjectsReq) -> list:
        """
        Search for data with a Filter (from ETSI ETSI EN 302 895 V1.1.1 (2014-09).

        Parameters
        ----------
        query : Filter
            Filter to be used for the search.
        """

    def insert(self, data: dict) -> int:
        """
        Insert data into the database, returns the index of the inserted data.

        Parameters
        ----------
        data : dict
            Data to be inserted into the database.
        """

    def get(self, index: int) -> dict:
        """
        Get data from the database with a given index.

        Parameters
        ----------
        index : int
            Index of the data to be retrieved.
        """

    def update(self, data: dict, index: int) -> bool:
        """
        Update data in the database with a given index, returns a boolean stating if update has been succesful.

        Parameters
        ----------
        data : dict
            Data to be updated in the database.
        index : int
            Index of the data to be updated.
        """

    def remove(self, data_object: dict = None) -> bool:
        """
        Remove data from the database with a given index, returns a boolean stating if removal has been succesful.

        Parameters
        ----------
        data_object : int
            Index of the data to be removed.
        """

    def all(self) -> list:
        """
        Get all data from the database.
        """

    def exists(self, field_name: str, data_object_id: int = None) -> bool:
        """
        Check if specific field exists.

        Key word field_name = "dataObjectID" will search for the "data_object_id" (index) in the database.

        Parameters
        ----------
        field_name : str
            Name of field to be checked
        Returns
        -------
        bool
            Indicates whether field exists
        """



# pylint: skip-file
ETSI_ITS_DSRC_ADDGRPC_ASN1_DESCRIPTIONS = """
--! @options: no-fields-header

ETSI-ITS-DSRC-AddGrpC {
  itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) ts103301 (103301) dsrc (6) addgrpc (0) major-version-2 (2) minor-version-1 (1)
}

DEFINITIONS AUTOMATIC TAGS::= BEGIN

IMPORTS

DeltaTime, FuelType, IntersectionID, LaneConnectionID, LaneID, NodeOffsetPointXY, NodeSetXY, PrioritizationResponseStatus, SignalGroupID, VehicleHeight
FROM ETSI-ITS-DSRC {
  itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) ts103301 (103301) dsrc (6) major-version-2 (2) minor-version-1 (1)
}

Altitude, DeltaAltitude, StationID, VehicleMass
FROM ETSI-ITS-CDD {
  itu-t (0) identified-organization (4) etsi (0) itsDomain (5) wg1 (1) 102894 cdd (2) major-version-4 (4) minor-version-2 (2)
};

/**
* This DF adds positioning support from the infrastructure to the vehicle.
*
* @field itsStationPositions: defines a list of ITS stations (e.g. vehicles) and their corresponding position on
*                             the driving lane as defined in the lane topology of the MapData message or the GNSS position
*                             deviation of the ITS Station from the high precision reference position in X/Y coordinates. It
*                             enables accurate, real-time positioning support to the moving ITS entities by the infrastructure.*
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
ConnectionManeuverAssist-addGrpC ::=	SEQUENCE {
  itsStationPosition ItsStationPositionList OPTIONAL,
  ...
}

/**
* This DF defines the trajectory for travelling through the conflict area of an intersection and connects 
* e.g an ingress with an egress lane. The trajectory is defined by two or more nodes. 
* The starting node overlaps e.g. with the node of the ingress lane towards the
* conflict zone. The ending node overlaps e.g. with the first node of the connected egress lane. 
* See the example in clause [ISO TS 19091] G.8.2.5.
*
* @field nodes: defines a list of nodes for the trajectory. It defines e.g. a geometric trajectory from an ingressing
*               to a connected egressing lane and the X/Y position value of the first node of the trajectory is the same as
*               the node of the ingress lane. The X/Y position of the last node is the same as the X/Y position of the first
*               node of the egressing lane.
* @field connectionID: defines the identifier of an allowed `maneuver` (e.g. ingress / egress relation). 
*               A generic Lane offers one or more allowed `maneuvers`, therefore the trajectory is reference to the related `maneuver`.
*
* @note: @ref Reg-GenericLane allows providing up to 4 connecting trajectories. In case a lane has more than 4 connecting trajectories,
* priority should be given to connecting trajectories of motorized traffic and complex manoeuvres.
*
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
ConnectionTrajectory-addGrpC ::= SEQUENCE {
  nodes         NodeSetXY,
  connectionID  LaneConnectionID,
  ...
}

/**
* This DF defines a list of prioritization responses e.g. public transport acceleration.
* The signal prioritization (e.g. public transport) procedure in this profile follows two strategies.
* - For simple prioritization requests, the CAM/SPAT messages are used. 
*   This allows the migration of old legal systems towards C-ITS.
*   In this case, the CAM message is used to trigger the request towards the traffic light controller. 
*   The traffic light controller checks the request and broadcasts the status for the priority request with this DF (see [ISO TS 19091] G.5.1.9).
* - For more complex signal requests, the SignalRequestMessage/SignalStatusMessage messages are to be used.
*
* @field activePrioritizations: list of Prioritizations.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
IntersectionState-addGrpC ::=	SEQUENCE {
  activePrioritizations  PrioritizationResponseList	OPTIONAL,
  ...
}

/**
* Lanes may have limitations regarding vehicle height (e.g. due to a tunnel) and vehicle weight (e.g. due to a bridge). 
*
* @field maxVehicleHeight: maximum allowed vehicle height
* @field maxVehicleWeight: maximum allowed vehicle mass
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
LaneAttributes-addGrpC ::= SEQUENCE {
  maxVehicleHeight  VehicleHeight OPTIONAL,
  maxVehicleWeight  VehicleMass   OPTIONAL,
  ...
}

/**
* This DF defines a list of three-dimensional positions of signal heads in an intersection. 
* It enables vehicles to identify the signal head location for optical evaluation of the traffic light. 
* Combined with the SPAT/MapData messages, it enables e.g. driving vehicles to enhance safety decision in critical situations.
*
* @field signalHeadLocations: list of geo positions
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
MapData-addGrpC ::=	SEQUENCE {
  signalHeadLocations  SignalHeadLocationList	OPTIONAL,
  ...
}

/**
* Priority and preemption have a considerable impact to the timing parameters in the SPAT message (eventState).
* User acceptance is expected to increase if the reason for sudden changes in timing parameters is communicated to them.
*
* @field stateChangeReason: reason code
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
MovementEvent-addGrpC ::= SEQUENCE {
  stateChangeReason  ExceptionalCondition OPTIONAL,
  ...
}

/**
* This DF defines additional attributes to support public transport and to enable a simple way of defining lane links.
*
* @field ptvRequest: defines control types attached to a node on a lane used by public transport for triggering
*                    the transmission of messages (e.g. prioritization request). It includes control points for public transport prioritization. 
*                    These control points are currently implemented by legacy systems using hardware sensors mounted on the roadside.
* @field nodeLink:   defines a link to one or to a set of another node/lane from this node. The nodeLink allows to set a link between specific nodes 
*                    of generic lanes or trajectories. This supports e.g. lane merging/diverging situations ([ISO TS 19091] G.8.2.7) and the linking of trajectories 
*                    in the conflict zone to lanes (see example [ISO TS 19091] G.8.2.5).
* @field node:       defines an identifier of this node.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
NodeAttributeSet-addGrpC  ::=	SEQUENCE {
  ptvRequest  PtvRequestType  OPTIONAL,
  nodeLink    NodeLink        OPTIONAL,
  node        Node            OPTIONAL,
  ...
}

/**
* This DF includes the altitude data element defined in the common data dictionary [ETSI CDD].
*
* @field elevation: the data element is replaced by the ETSI `altitude` data element using the regional extension. 
*                   The `altitude` data element is defined in Position3D-addGrpC of this profile.
*                   Position3D-addGrpC extends the @ref Position3D using the regional extension framework.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
Position3D-addGrpC ::=	SEQUENCE {
  altitude  Altitude,
  ...
}

/**
* This DF defines the driving restriction based on toxic emission type. 
* The meaning of the word `restriction` is ambiguous as it may have a double interpretation, being:
*  - only these vehicles are allowed OR 
*  - these vehicles are not allowed and all others are. 
* The former is what is intended by the base standard.
*
* @field emission: restriction baesed on emission.
* @field fuel: restriction baesed on fuel.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
RestrictionUserType-addGrpC ::=	SEQUENCE {
  emission  EmissionType OPTIONAL,
  fuel      FuelType     OPTIONAL,
  ...
}

/**
* Some road authorities like to give priority to vehicles based on the type of fuel they use. In addition,
* electric vehicles may receive priority based on their battery status.
*
* @field fuel: fuel used by vehicle.
* @field batteryStatus: current batter status of vehicle.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
RequestorDescription-addGrpC ::= SEQUENCE {
  fuel           FuelType      OPTIONAL,
  batteryStatus  BatteryStatus OPTIONAL,
  ...
}

/**
* The traffic control centre (TCC) may advice a public transport vehicle (e.g. bus) to synchronize his travel time. 
* This may happen when, for example, two busses, due to special traffic conditions, are out of schedule. 
* The first might be too late, the second too fast. The consequence is that the second is driving
* just behind the first and is empty as all passengers are within the first one. To avoid this often-occurring
* situation, the TCC transmits time synchronization advices to the public transport vehicles using the
* signal status message. 
*
* @field synchToSchedule: DeltaTime.
* @field rejectedReason: RejectedReason.
*
* @Note: The @ref PrioritizationResponseStatus provides optionally the reason for prioritization response rejection.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
SignalStatusPackage-addGrpC ::= SEQUENCE {
  synchToSchedule  DeltaTime      OPTIONAL,
  rejectedReason   RejectedReason OPTIONAL,
  ...
}

/**
* This DF is used to provide real-time positioning information feedback to a specific ITS station 
* (e.g. vehicle, pedestrian, bicycle) by infrastructure equipment.
*  The position information includes, for example, the driving, crossing lane and/or the X/Y coordinates in relation to
* the reference position of the MapData. The `timeReference` indicates the time stamp of the the
* message (received from an ITS station) for which the positioning feedback has been computed.
* 
* @field stationID: unique identifier.
* @field laneID: LaneID.
* @field nodeXY: NodeOffsetPointXY.
* @field timeReference: TimeReference.
*
* @note: The computation of the positioning feedback is out of focus of this standard.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
ItsStationPosition ::=	SEQUENCE {
  stationID      StationID,
  laneID         LaneID             OPTIONAL,
  nodeXY         NodeOffsetPointXY  OPTIONAL,
  timeReference  TimeReference      OPTIONAL,
  ...
}

ItsStationPositionList ::=	SEQUENCE SIZE(1..5) OF ItsStationPosition

/**
* This DF is used to to identify a node of a lane (waypoint) by its `lane` and node identifier `id`. 
*
* The `intersectionID` is used if the referenced lane belongs to an adjacent intersection. If the node
* belongs to a connection trajectory ([ISO TS 19091] G.5.1.2) the `connectionID` is used.
*
* @field id: unique identifier.
* @field lane: identifier from lane.
* @field connectionID: identifier from connection.
* @field intersectionID: identifier from intersection.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
Node ::= SEQUENCE {
  id              INTEGER,
  lane            LaneID           OPTIONAL,
  connectionID    LaneConnectionID OPTIONAL,
  intersectionID  IntersectionID   OPTIONAL,
  ...
}

NodeLink ::= SEQUENCE SIZE (1..5) OF Node

/**
* This DF is used to provide the prioritization status response and the
* signal group identifier for a specific ITS station (e.g. vehicle).
*
* @field stationID: StationID.
* @field priorState: PrioritizationResponseStatus.
* @field signalGroup: SignalGroupID.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
PrioritizationResponse ::=	SEQUENCE {
  stationID    StationID,
  priorState   PrioritizationResponseStatus,
  signalGroup  SignalGroupID,
  ...
}

PrioritizationResponseList ::=	SEQUENCE SIZE(1..10) OF PrioritizationResponse

/**
* This DF defines the XYZ position of a signal head within an intersection
* and indicates the related signal group identifier.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
SignalHeadLocation ::=	SEQUENCE {
  nodeXY         NodeOffsetPointXY,
  nodeZ          DeltaAltitude,
  signalGroupID  SignalGroupID,
  ...
}

SignalHeadLocationList ::=	SEQUENCE (SIZE(1..64)) OF	SignalHeadLocation

/**
* This DE defines an enumerated list of battery states.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
BatteryStatus ::= ENUMERATED {
  unknown,
  critical,
  low,
  good,
  ...
}

/**
* This DE defines an enumerated list of toxic emission types for vehicles.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
EmissionType ::= ENUMERATED {
  euro1,
  euro2,
  euro3,
  euro4,
  euro5,
  euro6,
  ...
}

/**
* This DE defines a list of reasons for sudden changes in
* eventState parameters, thereby offering a reason for extended waiting times.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
ExceptionalCondition ::= ENUMERATED {
  unknown,
  publicTransportPriority,
  emergencyVehiclePriority,
  trainPriority,
  bridgeOpen,
  vehicleHeight,
  weather,
  trafficJam,
  tunnelClosure,
  meteringActive,
  truckPriority,
  bicyclePlatoonPriority,
  vehiclePlatoonPriority,
  ...
}

/**
* This DE defines a list of activation requests used for C-ITS migration of legacy public 
* transport prioritization systems. 
* The activation points are used while approaching to an intersection.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
PtvRequestType  ::= ENUMERATED {
  preRequest,
  mainRequest,
  doorCloseRequest,
  cancelRequest,
  emergencyRequest,
  ...
}

/**
* This DE defines a list of reasons for rejected priority requests.
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
RejectedReason ::= ENUMERATED {
  unknown,
  exceptionalCondition,
  maxWaitingTimeExceeded,
  ptPriorityDisabled,
  higherPTPriorityGranted,
  vehicleTrackingUnknown,
  ...
}

/**
* This DE defines a value in milliseconds in the current minute related to UTC time. 
* The range of 60 000 covers one minute (60 seconds * 1 000 milliseconds)
*
* @category: Infrastructure information
* @revision: V1.3.1
*/
TimeReference ::= INTEGER { oneMilliSec(1) } (0..60000)

/** 
* ## References:
* 1. [ISO TS 19091]: "Intelligent transport systems - Cooperative ITS - Using V2I and I2V communications for applications related to signalized intersections".
* 2. [SAE J2735]: "SURFACE VEHICLE STANDARD - V2X Communications Message Set Dictionary"
*/

END
"""



from __future__ import annotations
import time
import hashlib
from .gbc_extended_header import GBCExtendedHeader
from .gn_address import GNAddress
from .mib import MIB
from .position_vector import LongPositionVector, TST
from .exceptions import DuplicatedPacketException, IncongruentTimestampException


class LocationTableEntry:
    """
    Location table entry class. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 8.1.2


    Attributes
    ----------
    mib : MIB
        MIB to use.
    position_vector : LongPositionVector
        Position vector of the ITS-S.
    ls_pending : bool
        Flag indicating that a Location Service (LS) (clause 10.2.4) is in progress.
    is_neighbour : bool
        Flag indicating that the GeoAdhoc router is in direct communication
        range, i.e. is a neighbour.
    dpl : List[bytes]
        Duplicate packet list for source GN_ADDR. (List of hashes of the packets)
    tst : TST
        Timestamp TST(GN_ADDR): The timestamp of the last packet from the source GN_ADDR that was identified
        as 'not duplicated'
    pdr : int
        Packet data rate PDR(GN_ADDR) as Exponential Moving Average (EMA) (clause B.2).
    """

    def __init__(self, mib: MIB):
        self.mib = mib
        # In the future the version will be corrected, now if the version is not the same all packets are dropped
        self.version: int = mib.itsGnProtocolVersion
        self.position_vector: LongPositionVector = LongPositionVector()
        self.ls_pending: bool = False
        self.is_neighbour: bool = False
        self.dpl: list[bytes] = []
        self.tst: TST = TST()
        self.pdr: int = 0

    def get_gn_address(self) -> GNAddress:
        """
        Get the GN address.

        Returns
        -------
        GNAddress
            GN address.
        """
        return self.position_vector.gn_addr

    def update_position_vector(self, position_vector: LongPositionVector) -> None:
        """
        Updates the position vector.
        Annex C2 of ETSI EN 302 636-4-1 V1.4.1 (2020-01)
        The algorithm is implemented partially on the TST

        Parameters
        ----------
        position_vector : LongPositionVector
            Position vector to update.

        Raises
        ------
        IncongruentTimestampException
            If there has been another packet with posterior timestamp received before.
        """
        if self.position_vector.tst.msec == 0:
            self.position_vector = position_vector
        elif position_vector.tst > self.position_vector.tst:
            self.position_vector = position_vector
        else:
            raise IncongruentTimestampException("Position vector not updated")

    def update_pdr(self, position_vector: LongPositionVector, packet_size: int) -> None:
        """
        Updates the Packet Data Rate (PDR).
        Annex B2 of ETSI EN 302 636-4-1 V1.4.1 (2020-01)

        Parameters
        ----------
        position_vector : LongPositionVector
            Position vector of the packet.
        packet_size : int
            Size of the packet.
        """
        time_since_last_update = (position_vector.tst - self.tst) / 1000
        self.tst = position_vector.tst
        if time_since_last_update > 0:
            current_pdr = packet_size / time_since_last_update
            # Equation B1
            beta = self.mib.itsGnMaxPacketDataRateEmaBeta / 100
            self.pdr = beta * self.pdr + (1 - beta) * current_pdr

    def update_with_shb_packet(
        self, position_vector: LongPositionVector, packet: bytes
    ) -> None:
        """
        Updates the entry with a SHB packet.

        Follows the steps 4, 5 and 6 of the algorithm in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 10.3.10.3

        Parameters
        ----------
        position_vector : LongPositionVector
            Position vector of the packet.
        packet : bytes
            SHB packet (without the basic header, the common header and the position vector).

        Raises
        ------
        IncongruentTimestampException
            If there has been another packet with posterior timestamp received before.
        DuplicatedPacketException
            If the packet is duplicated.
        """
        # Own added check (not in the standard)
        self.check_duplicate_packet(packet)
        # End own added check
        # step 4
        self.update_position_vector(position_vector)
        # step 5
        self.update_pdr(position_vector, (len(packet) + 8 + 4))
        # step 6
        self.is_neighbour = True

    def update_with_gbc_packet(
        self, packet: bytes, gbc_extended_header: GBCExtendedHeader
    ) -> None:
        """
        Updates the entry with a SHB packet.

        Follows the steps 3, 4, 5 and 6 of the algorithm in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 10.3.11.3

        Parameters
        ----------
        packet : bytes
            GBC packet (without the basic header, the common header and the extended header).
        gbc_extended_header : GBCExtendedHeader
            GBC extended header.

        Raises
        ------
        IncongruentTimestampException
            If there has been another packet with posterior timestamp received before.
        DuplicatedPacketException
            If the packet is duplicated.
        """
        position_vector = gbc_extended_header.so_pv
        # Own added check (not in the standard)
        self.check_duplicate_packet(packet)
        # End own added check
        # step 4
        self.update_position_vector(position_vector)
        # step 5
        self.update_pdr(position_vector, (len(packet) + 8 + 4))
        # step 6
        self.is_neighbour = False

    def check_duplicate_packet(self, packet: bytes) -> None:
        """
        Checks if the packet is duplicated.

        ETSI EN 302 636-4-1 V1.4.1 (2020-01)

        Temporary implementation. The DPL is not implemented yet.

        Parameters
        ----------
        packet : bytes
            Packet to check.

        Raises
        ------
        DuplicatedPacketException
            If the packet is duplicated.
        """
        packet_hash = hashlib.sha256(packet).digest()
        if packet_hash in self.dpl:
            raise DuplicatedPacketException("Packet is duplicated")
        self.dpl.append(packet_hash)
        if len(self.dpl) > 50:
            self.dpl.pop(0)


class LocationTable:
    """
    Location table class.  ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 8.1.1

    Attributes
    ----------
    mib : MIB
        MIB to use.
    loc_t : List[LocationTableEntry]
        Location table.
    """

    def __init__(self, mib: MIB):
        """
        Constructor.

        Parameters
        ----------
        mib : MIB
            MIB to use.
        """
        self.mib = mib
        self.loc_t: list[LocationTableEntry] = []

    def get_entry(self, gn_address: GNAddress) -> LocationTableEntry:
        """
        Gets the entry of the location table.

        Parameters
        ----------
        gn_address : GNAddress
            GN address.

        Returns
        -------
        LocationTableEntry
            Location table entry.
        """
        for entry in self.loc_t:
            if entry.get_gn_address() == gn_address:
                return entry
        return None

    def refresh_table(self) -> None:
        """
        Removes the entries that have expired.

        Temporarily solution following ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 8.1.3
        """
        current_time = TST()
        current_time.set_in_normal_timestamp_seconds(time.time())
        for entry in self.loc_t:
            if (current_time - entry.position_vector.tst) > self.mib.itsGnLifetimeLocTE:
                self.loc_t.remove(entry)

    def new_shb_packet(
        self, position_vector: LongPositionVector, packet: bytes
    ) -> None:
        """
        Updates the location table with a new packet.

        Parameters
        ----------
        position_vector : LongPositionVector
            Position vector of the packet.
        packet : bytes
            SHB packet (without the basic header, the common header and the position vector).

        Raises
        ------
        IncongruentTimestampException
            If there has been another packet with posterior timestamp received before.
        DuplicatedPacketException
            If the packet is duplicated.
        """
        entry: LocationTableEntry = self.get_entry(position_vector.gn_addr)
        if entry:
            entry.update_with_shb_packet(position_vector, packet)
        else:
            entry = LocationTableEntry(self.mib)
            entry.update_with_shb_packet(position_vector, packet)
            self.loc_t.append(entry)
        self.refresh_table()

    def new_gbc_packet(
        self, gbc_extended_header: GBCExtendedHeader, packet: bytes
    ) -> None:
        """
        Updates the location table with a new packet.

        Parameters
        ----------
        gbc_extended_header : GBCExtendedHeader
            GBC extended header.
        packet : bytes
            GBC packet (without the basic header, the common header and the extended header).

        Raises
        ------
        IncongruentTimestampException
            If there has been another packet with posterior timestamp received before.
        DuplicatedPacketException
            If the packet is duplicated.
        """
        entry: LocationTableEntry = self.get_entry(gbc_extended_header.so_pv.gn_addr)
        if entry:
            entry.update_with_gbc_packet(packet, gbc_extended_header)
        else:
            entry = LocationTableEntry(self.mib)
            entry.update_with_gbc_packet(packet, gbc_extended_header)
            self.loc_t.append(entry)
        self.refresh_table()

    def get_neighbours(self) -> list[LocationTableEntry]:
        """
        Gets the neighbours.

        Returns
        -------
        List[LocationTableEntry]
            List of neighbours.
        """
        neighbours: list[LocationTableEntry] = []
        for entry in self.loc_t:
            if entry.is_neighbour:
                neighbours.append(entry)
        return neighbours



from __future__ import annotations
from collections.abc import Callable
from enum import Enum
import math
from ..linklayer.exceptions import (
    SendingException,
    PacketTooLongException,
)
from .mib import (
    MIB,
    LocalGnAddrConfMethod,
    NonAreaForwardingAlgorithm,
    AreaForwardingAlgorithm,
)
from .gn_address import GNAddress
from .service_access_point import (
    HeaderType,
    TopoBroadcastHST,
    GeoBroadcastHST,
    GNDataRequest,
    ResultCode,
    GNDataConfirm,
    GNDataIndication,
    Area,
    PacketTransportType,
)
from .basic_header import BasicNH, BasicHeader
from .common_header import CommonHeader
from .gbc_extended_header import GBCExtendedHeader
from .position_vector import LongPositionVector
from .location_table import LocationTable
from ..security.sign_service import SignService
from ..security.security_profiles import SecurityProfile
from ..security.sn_sap import SNSIGNConfirm, SNSIGNRequest
from .exceptions import (
    DADException,
    DecapError,
    DecodeError,
    DuplicatedPacketException,
    IncongruentTimestampException,
)

EARTH_RADIUS = 6371000  # Radius of the Earth in meters


class GNForwardingAlgorithmResponse(Enum):
    """
    GN Forwarding Algorithm Selection Response. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex D.

    Attributes
    ----------
    AREA-FORWARDING : 1
        Area Forwarding.
    NON-AREA-FORWARDING : 2
        Non-Area Forwarding.
    DISCARTED : 3
        Discarted.
    """

    AREA_FORWARDING = 1
    NON_AREA_FORWARDING = 2
    DISCARTED = 3


class Router:
    """
    Geonetworking Router

    Handles the routing of Geonetworking packets. As specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01).

    """

    def __init__(self, mib: MIB, sign_service: SignService = None) -> None:
        """
        Initialize the router.

        Parameters
        ----------
        mib : MIB
            MIB to use.
        """
        self.mib = mib
        self.gn_address = GNAddress()
        self.setup_gn_address()
        self.ego_position_vector = LongPositionVector()
        self.ego_position_vector.set_gn_addr(self.gn_address)
        self.link_layer = None
        self.location_table = LocationTable(mib)
        self.sign_service: SignService = sign_service
        self.indication_callback = None
        self.sequence_number = 0

    def get_sequence_number(self) -> int:
        """
        Get the current sequence number.

        Returns
        -------
        int
            Current sequence number.
        """
        self.sequence_number = (self.sequence_number + 1) % (2**16 - 1)
        return self.sequence_number

    def register_indication_callback(
        self, callback: Callable[[GNDataIndication], None]
    ) -> None:
        """
        Registers a callback for GNDataIndication.

        Parameters
        ----------
        callback : Callable[[GNDataIndication], None]
            Callback to register.
        """
        self.indication_callback = callback

    def setup_gn_address(self) -> None:
        # pylint: disable=no-else-raise
        """
        Set the GN address of the router.

        Raises
        ------
        NotImplementedError :
            If the local GN address configuration method is not implemented.
        """
        if self.mib.itsGnLocalGnAddrConfMethod == LocalGnAddrConfMethod.MANAGED:
            raise NotImplementedError(
                "Managed GN address configuration is not implemented."
            )
        elif self.mib.itsGnLocalGnAddrConfMethod == LocalGnAddrConfMethod.AUTO:
            self.gn_address = self.mib.itsGnLocalGnAddr
        elif self.mib.itsGnLocalGnAddrConfMethod == LocalGnAddrConfMethod.ANONYMOUS:
            raise NotImplementedError(
                "Anonymous GN address configuration is not implemented."
            )

    def gn_data_request_shb(self, request: GNDataRequest) -> GNDataConfirm:
        """
        Handle a Single Hop Broadcast GNDataRequest.

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to handle.
        """
        basic_header = BasicHeader()
        basic_header.initialize_with_mib(self.mib)
        basic_header.set_rhl(1)
        common_header = CommonHeader()
        common_header.initialize_with_request(request)
        long_position_vector = self.ego_position_vector
        media_dependant_data = b"\x00\x00\x00\x00"
        packet = b""
        if request.security_profile == SecurityProfile.COOPERATIVE_AWARENESS_MESSAGE:
            if self.sign_service is None:
                raise NotImplementedError("Security profile not implemented")
            common_header = CommonHeader()
            common_header.initialize_with_request(request)
            long_position_vector = self.ego_position_vector
            media_dependant_data = b"\x00\x00\x00\x00"
            tbs_packet = (
                common_header.encode_to_bytes()
                + long_position_vector.encode()
                + media_dependant_data
                + request.data
            )
            sign_request = SNSIGNRequest(
                tbs_message_length=len(tbs_packet),
                tbs_message=tbs_packet,
                its_aid=request.its_aid,
                permissions=request.security_permissions,
                permissions_length=len(request.security_permissions),
            )
            sign_confirm: SNSIGNConfirm = self.sign_service.sign_cam(sign_request)
            basic_header.set_nh(BasicNH.SECURED_PACKET)
            packet = basic_header.encode_to_bytes() + sign_confirm.sec_message

        else:
            packet = (
                basic_header.encode_to_bytes()
                + common_header.encode_to_bytes()
                + long_position_vector.encode()
                + media_dependant_data
                + request.data
            )
        confirm = GNDataConfirm()

        try:
            if self.link_layer:
                self.link_layer.send(packet)
        except PacketTooLongException:
            confirm.result_code = ResultCode.MAXIMUM_LENGTH_EXCEEDED
        except SendingException:
            confirm.result_code = ResultCode.UNSPECIFIED

        confirm.result_code = ResultCode.ACCEPTED

        return confirm

    @staticmethod
    def calculate_distance(
        coord1: tuple[float, float], coord2: tuple[float, float]
    ) -> tuple[float, float]:
        """
        Returns the distance between two coordinates in meters.
        As specified in ETSI EN 302 931 - V1.0.0
        Latitude -> x
        Longitude -> -y

        Returns
        -------
        Tuple[float, float]:
            Tuple of x distance, y distance
        """
        lat1, lon1 = coord1
        lat2, lon2 = coord2

        # Convert latitude and longitude to radians
        lat1 = math.radians(lat1)
        lon1 = math.radians(lon1)
        lat2 = math.radians(lat2)
        lon2 = math.radians(lon2)

        # Calculate the differences in latitude and longitude
        dlat = lat2 - lat1
        dlon = lon2 - lon1

        # Calculate the distance along the y-axis (-longitude)
        y_distance = EARTH_RADIUS * dlon * math.cos((lat1 + lat2) / 2)

        # Calculate the distance along the x-axis (latitude)
        x_distance = (-1) * EARTH_RADIUS * dlat

        return x_distance, y_distance

    @staticmethod
    def transform_distance_angle(
        distance: tuple[float, float], angle: int
    ) -> tuple[float, float]:
        """
        Adapts the X,Y pointed at north to the right angle

        Returns
        -------
        tuple[float, float]
            X and Y distances adapted to the angle
        """
        n_angle = math.radians(angle)
        new_x_distance = math.cos(n_angle) * distance[0]
        new_y_distance = math.sin(n_angle) * distance[1]
        return (new_x_distance, new_y_distance)

    def gn_geometric_function_f(
        self, area_type: GeoBroadcastHST, area: Area, lat: int, lon: int
    ) -> float:
        """
        Implements the Geometric function F to determine spatial characteristics of a point P(x,y).
        As specified in EN 302 931 - V1.0.0 Section 5

        Parameters
        ----------
        area_type : GeoBroadcastHST
            Type of the area.
        area : Area
            Area of the circle.
        lat : int
            Latitude of the point P. In 1/10 microdegrees.
        lon : int
            Longitude of the point P. In 1/10 microdegrees.
        """
        coord1 = (area.latitude / 10000000, area.longitude / 10000000)
        coord2 = (lat / 10000000, lon / 10000000)
        x_distance, y_distance = Router.calculate_distance(coord1, coord2)
        if area_type == GeoBroadcastHST.GEOBROADCAST_CIRCLE:
            return 1 - (x_distance / area.a) ** 2 - (y_distance / area.a) ** 2
        if area_type == GeoBroadcastHST.GEOBROADCAST_ELIP:
            return 1 - (x_distance / area.a) ** 2 - (y_distance / area.b) ** 2
        if area_type == GeoBroadcastHST.GEOBROADCAST_RECT:
            return min(1 - (x_distance / area.a) ** 2, (y_distance / area.b) ** 2)
        raise ValueError("Invalid area type")

    def gn_forwarding_algorithm_selection(
        self, request: GNDataRequest
    ) -> GNForwardingAlgorithmResponse:
        """

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to handle.
        """
        result = self.gn_geometric_function_f(
            request.packet_transport_type.header_subtype,
            request.area,
            self.ego_position_vector.latitude,
            self.ego_position_vector.longitude,
        )

        if result >= 0:
            return GNForwardingAlgorithmResponse.AREA_FORWARDING
        # TODO: Parts of the forwarding algorithm selection
        return GNForwardingAlgorithmResponse.DISCARTED

    def gn_data_forward_gbc(
        self,
        basic_header: BasicHeader,
        common_header: CommonHeader,
        gbc_extended_header: GBCExtendedHeader,
        packet: bytes,
    ) -> GNDataConfirm:
        """
        Function called when a GBC packet has to be fowraded.

        Parameters
        ----------
        basic_header : BasicHeader
            Basic header of the packet.
        common_header : CommonHeader
            Common header of the packet.
        gbc_extended_header : GBCExtendedHeader
            Extended header of the packet.
        packet : bytes
            Packet to forward. (Without headers)
        """
        confirm = GNDataConfirm()
        # TODO: Location Service (LS) packet buffers (step 8)
        basic_header.set_rhl(basic_header.rhl - 1)
        # 10) if no neighbour exists, i.e. the LocT does not contain a LocTE with the IS_NEIGHBOUR flag set to TRUE,
        # and SCF for the traffic class in the TC field of the Common Header is set, buffer the GBC packet in the BC
        # forwarding packet buffer and omit the execution of further steps;
        if len(self.location_table.get_neighbours()) > 0 or not common_header.tc.scf:
            # 11) execute the forwarding algorithm procedures (starting with annex D);
            request = GNDataRequest()
            request.area = Area()
            request.area.latitude = gbc_extended_header.latitude
            request.area.longitude = gbc_extended_header.longitude
            request.area.a = gbc_extended_header.a
            request.area.b = gbc_extended_header.b
            request.area.angle = gbc_extended_header.angle
            request.packet_transport_type.header_subtype = common_header.hst
            algorithm = self.gn_forwarding_algorithm_selection(request)
            # 12) if the return value of the forwarding algorithm is 0 (packet is buffered in a forwarding packet
            # buffer) or -1 (packet is discarded), omit the execution of further steps;
            if algorithm == GNForwardingAlgorithmResponse.AREA_FORWARDING:
                # TODO: step 13
                # 14) pass the GN-PDU to the LL protocol entity via the IN interface and set the destination
                # address to the LL address of the next hop LL_ADDR_NH.
                packet: bytes = (
                    basic_header.encode_to_bytes()
                    + common_header.encode_to_bytes()
                    + gbc_extended_header.encode()
                    + packet
                )
                try:
                    if self.link_layer:
                        self.link_layer.send(packet)
                except PacketTooLongException:
                    confirm.result_code = ResultCode.MAXIMUM_LENGTH_EXCEEDED
                except SendingException:
                    confirm.result_code = ResultCode.UNSPECIFIED

                confirm.result_code = ResultCode.ACCEPTED
        else:
            packet: bytes = (
                basic_header.encode_to_bytes()
                + common_header.encode_to_bytes()
                + gbc_extended_header.encode()
                + packet
            )
            try:
                if self.link_layer:
                    self.link_layer.send(packet)
            except PacketTooLongException:
                confirm.result_code = ResultCode.MAXIMUM_LENGTH_EXCEEDED
            except SendingException:
                confirm.result_code = ResultCode.UNSPECIFIED

            confirm.result_code = ResultCode.ACCEPTED

        return confirm

    def gn_data_request_gbc(self, request: GNDataRequest) -> GNDataConfirm:
        """
        Handle a Geo Broadcast GNDataRequest.

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to handle.
        """
        confirm = GNDataConfirm()

        # 1) create a GN-PDU with the T/GN6-SDU as payload and a GBC packet header (clause 9.8.5):
        #   a) set the fields of the Basic Header (clause 10.3.2);
        basic_header = BasicHeader()
        basic_header.initialize_with_mib(self.mib)
        #   b) set the fields of the Common Header (clause 10.3.4);
        common_header = CommonHeader()
        common_header.initialize_with_request(request)
        common_header.ht = HeaderType.GEOBROADCAST
        common_header.hst = request.packet_transport_type.header_subtype
        #   c) set the fields of the GBC Extended Header (table 36);
        geo_broadcast_extended_header = GBCExtendedHeader()
        geo_broadcast_extended_header.initialize_with_request(request)
        geo_broadcast_extended_header.sn = self.get_sequence_number()
        geo_broadcast_extended_header.so_pv = self.ego_position_vector
        # 2) if no neighbour exists, i.e. the LocT does not contain a LocTE with the IS_NEIGHBOUR flag set to TRUE,
        # and SCF for the traffic class in the service primitive GN-DATA.request parameter Traffic class is enabled,
        # then buffer the GBC packet in the BC forwarding packet buffer and omit the execution of further steps;
        if (
            len(self.location_table.get_neighbours()) > 0
            or not request.traffic_class.scf
        ):
            # 3) execute the forwarding algorithm procedures (starting with annex D);
            algorithm = self.gn_forwarding_algorithm_selection(request)
            # 4) if the return value of the forwarding algorithm is 0 (packet is buffered in the BC forwarding packet
            # buffer or in the CBF buffer) or -1 (packet is discarded), omit the execution of further steps;
            if algorithm == GNForwardingAlgorithmResponse.AREA_FORWARDING:
                # TODO: steps 5-7
                # 8) pass the GN-PDU to the LL protocol entity via the IN interface and set the destination address to
                # the LL address of the next hop LL_ADDR_NH.
                packet: bytes = (
                    basic_header.encode_to_bytes()
                    + common_header.encode_to_bytes()
                    + geo_broadcast_extended_header.encode()
                    + request.data
                )
                try:
                    if self.link_layer:
                        self.link_layer.send(packet)
                except PacketTooLongException:
                    confirm.result_code = ResultCode.MAXIMUM_LENGTH_EXCEEDED
                except SendingException:
                    confirm.result_code = ResultCode.UNSPECIFIED

                confirm.result_code = ResultCode.ACCEPTED
        else:
            packet: bytes = (
                basic_header.encode_to_bytes()
                + common_header.encode_to_bytes()
                + geo_broadcast_extended_header.encode()
                + request.data
            )
            try:
                if self.link_layer:
                    self.link_layer.send(packet)
            except PacketTooLongException:
                confirm.result_code = ResultCode.MAXIMUM_LENGTH_EXCEEDED
            except SendingException:
                confirm.result_code = ResultCode.UNSPECIFIED

            confirm.result_code = ResultCode.ACCEPTED

        return confirm

    def gn_data_request(self, request: GNDataRequest) -> GNDataConfirm:
        """
        Handle a GNDataRequest.

        Parameters
        ----------
        request : GNDataRequest
            GNDataRequest to handle.

        Raises
        ------
        NotImplementedError : PacketTransportType not implemented

        Returns
        -------
        GNDataConfirm :
            Confirmation of the process of the packet.
        """
        if (request.packet_transport_type.header_type == HeaderType.TSB) and (
            request.packet_transport_type.header_subtype == TopoBroadcastHST.SINGLE_HOP
        ):
            return self.gn_data_request_shb(request)
        if request.packet_transport_type.header_type == HeaderType.GEOBROADCAST:
            return self.gn_data_request_gbc(request)
        raise NotImplementedError("PacketTransportType not implemented")

    def gn_data_indicate_shb(
        self, packet: bytes, common_header: CommonHeader
    ) -> GNDataIndication:
        """
        Handle a Single Hop Broadcast GeoNetworking packet.

        Parameters
        ----------
        packet : bytes
            GeoNetworking packet to handle.
        common_header : CommonHeader
            CommonHeader of the packet.
        """
        # ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section
        indication = GNDataIndication()
        try:
            long_position_vector = LongPositionVector()
            long_position_vector.decode(packet[0:24])
            packet = packet[24:]
            # Ignore Media Dependant Data
            packet = packet[4:]
            self.location_table.new_shb_packet(long_position_vector, packet)
            indication.upper_protocol_entity = common_header.nh
            indication.source_position_vector = long_position_vector
            indication.traffic_class = common_header.tc
            indication.length = len(packet)
            indication.data = packet
        except DADException:
            print("Duplicate Address Detected!")
        except IncongruentTimestampException:
            print("Incongruent Timestamp Detected!")
        except DuplicatedPacketException:
            print("Packet is duplicated")
        except DecodeError as e:
            print(str(e))
        return indication

    def gn_data_indicate_gbc(
        self, packet: bytes, common_header: CommonHeader
    ) -> GNDataIndication:
        """
        Handle a GeobroadcastBroadcast GeoNetworking packet.

        Parameters
        ----------
        packet : bytes
            GeoNetworking packet to handle (without the basic header and common header)
        common_header : CommonHeader
            CommonHeader of the packet.
        """
        indication = GNDataIndication()
        gbc_extended_header = GBCExtendedHeader()
        gbc_extended_header.decode(packet[0:44])
        packet = packet[44:]
        area = Area()
        area.a = gbc_extended_header.a
        area.b = gbc_extended_header.b
        area.latitude = gbc_extended_header.latitude
        area.longitude = gbc_extended_header.longitude
        area.angle = gbc_extended_header.angle
        area_f = self.gn_geometric_function_f(
            common_header.hst,
            area,
            self.ego_position_vector.latitude,
            self.ego_position_vector.longitude,
        )
        if area_f < 0 and (
            self.mib.itsGnNonAreaForwardingAlgorithm
            in (
                NonAreaForwardingAlgorithm.GREEDY,
                NonAreaForwardingAlgorithm.UNSPECIFIED,
            )
        ):
            pass
        elif area_f >= 0 and (
            self.mib.itsGnAreaForwardingAlgorithm
            in (AreaForwardingAlgorithm.UNSPECIFIED, AreaForwardingAlgorithm.SIMPLE)
        ):
            pass
        try:
            self.duplicate_address_detection(gbc_extended_header.so_pv.gn_addr)
            self.location_table.new_gbc_packet(gbc_extended_header, packet)
            if area_f >= 0:
                indication.upper_protocol_entity = common_header.nh
                indication.packet_transport_type = PacketTransportType()
                indication.packet_transport_type.header_type = HeaderType.GEOBROADCAST
                indication.packet_transport_type.header_subtype = common_header.hst
                indication.source_position_vector = gbc_extended_header.so_pv
                indication.traffic_class = common_header.tc
                # TODO: Extend the indication information
                # indication.length = len(packet[44:])
                # indication.data = packet[44:]
                indication.length = len(packet)
                indication.data = packet
        except DADException:
            print("Duplicate Address Detected!")
        except IncongruentTimestampException:
            print("Incongruent Timestamp Detected!")
        except DecodeError as e:
            print(str(e))
        return indication

    def gn_data_indicate(self, packet: bytes) -> None:
        # pylint: disable=no-else-raise, too-many-branches
        """
        Method to indicate a GeoNetworking packet.

        Lower level layers should call this method to indicate a GeoNetworking packet.

        Parameters
        ----------
        packet : bytes
            GeoNetworking packet to indicate.

        Raises
        ------
        NotImplementedError : Version not implemented
        """
        indication = GNDataIndication()
        # ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 10.3.3
        # Decap the common header
        basic_header = BasicHeader()
        basic_header.decode_from_bytes(packet[0:4])
        packet = packet[4:]
        if basic_header.version != self.mib.itsGnProtocolVersion:
            raise NotImplementedError("Version not implemented")
        if basic_header.nh == BasicNH.COMMON_HEADER:
            # ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 10.3.5
            # Decap the common header
            common_header = CommonHeader()
            common_header.decode_from_bytes(packet[0:8])
            packet = packet[8:]
            if basic_header.rhl > self.mib.itsGnDefaultHopLimit:
                raise DecapError("Hop limit exceeded")
            # TODO: Forwarding packet buffer flush
            if common_header.ht == HeaderType.ANY:
                raise NotImplementedError("Any packet (Common Header) not implemented")
            elif common_header.ht == HeaderType.BEACON:
                raise NotImplementedError("Beacon not implemented")
            elif common_header.ht == HeaderType.GEOUNICAST:
                raise NotImplementedError("Geounicast not implemented")
            elif common_header.ht == HeaderType.GEOANYCAST:
                raise NotImplementedError("Geoanycast not implemented")
            elif common_header.ht == HeaderType.GEOBROADCAST:
                indication = self.gn_data_indicate_gbc(packet, common_header)
            elif common_header.ht == HeaderType.TSB:
                if common_header.hst == TopoBroadcastHST.SINGLE_HOP:
                    indication = self.gn_data_indicate_shb(packet, common_header)
                else:
                    raise NotImplementedError("TopoBroadcast not implemented")
            elif common_header.ht == HeaderType.LS:
                raise NotImplementedError("Location Service not implemented")
            else:
                raise NotImplementedError("Any packet (Common Header) not implemented")

        elif basic_header.nh == BasicNH.SECURED_PACKET:
            raise NotImplementedError("Secured packet not implemented")
        else:
            raise NotImplementedError("ANY next header not implemented")
        if self.indication_callback:
            self.indication_callback(indication)

    def duplicate_address_detection(self, gn_addr: GNAddress) -> None:
        """
        Perform Duplicate Address Detection (DAD) on the given GNAddress.
        Specified in ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 10.2.1.5

        Parameters
        ----------
        gn_addr : GNAddress
            GNAddress to perform Duplicate Address Detection on.
        """
        if self.mib.itsGnLocalGnAddr == gn_addr:
            raise DADException("Duplicate Address Detected!")
            # TODO : Handle the reset of the GN address as said in the standard

    def refresh_ego_position_vector(self, tpv: dict) -> None:
        """
        Refresh the ego position vector.
        """
        self.ego_position_vector.refresh_with_tpv_data(tpv)



"""
Location Service Module

This module contains the LocationService class, which is used to get the current
location from a GPSD server.

Usage:
    from location_service import LocationService

    # Initialize a new LocationService instance
    location_service = LocationService(gpsd_host='localhost', gpsd_port=2947)

    # Add a callback function to be called when a new location is received
    def my_callback(location):
        print(f'New location received: {location}')

    location_service.add_callback(my_callback)

Attributes
----------
gpsd_host : str
    The hostname or IP address of the GPSD server.
gpsd_port : int
    The port number of the GPSD server.
socket : socket.socket
    The socket object used to communicate with the GPSD server.
callbacks : List[Callable[[dict], None]]
    A list of callback functions to be called when a new location is received.

Classes
-------
LocationService
    A class that connects to a GPSD server and provides current location data.
"""
import threading
import socket
import json
from .location_service import LocationService


class GPSDLocationService(LocationService):
    """
    Location Service

    This class is used to get the location.

    Attributes
    ----------
    gpsd_host : str
        The host of the gpsd server.
    gpsd_port : int
        The port of the gpsd server.
    socket : socket.socket
        The socket to the gpsd server.
    callbacks : List[Callable[[dict], None]]
        The callbacks to call when a new location is received.
    """

    def __init__(self, gpsd_host: str = "localhost", gpsd_port: int = 2947) -> None:
        """
        Initialize the Location Service.

        Parameters
        ----------
        gpsd_host : str
            The host of the gpsd server.
        gpsd_port : int
            The port of the gpsd server.
        """
        super().__init__()
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.gpsd_host = gpsd_host
        self.gpsd_port = gpsd_port
        self.stop_event = threading.Event()
        self.location_service_thread = threading.Thread(target=self.start, daemon=True)
        self.location_service_thread.start()

    def __del__(self) -> None:
        """
        Destroy the Location Service.
        """
        self.socket.close()

    def reconnect(self) -> None:
        """
        Reconnect the Location Service.
        """
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket.setdefaulttimeout(10)
        connected = False
        while not connected:
            try:
                self.socket.connect((self.gpsd_host, self.gpsd_port))
                connected = True
            except TimeoutError:
                pass
            except InterruptedError:
                pass
            except ConnectionRefusedError:
                print("ConnectionRefusedError: [Errno 111] Connection refused, trying again...")
        try:
            self.socket.send(b'?WATCH={"enable":true,"json":true};')
        except BrokenPipeError:
            self.reconnect()

    def start(self) -> None:
        """
        Starts the Location Service Thread.
        """
        self.reconnect()
        while not self.stop_event.is_set():
            try:
                data = self.socket.recv(1024)
                if data:
                    json_data = {}
                    try:
                        json_data = json.loads(data)
                    except ValueError:
                        print("Could not decode data")
                        print(data)
                    if "class" in json_data and json_data["class"] == "TPV":
                        self.send_to_callbacks(json_data)
            except BrokenPipeError:
                self.reconnect()
            except ConnectionResetError:
                self.reconnect()
            except TimeoutError:
                self.reconnect()



from __future__ import annotations
import os

import tinydb

from .ldm_constants import OPERATOR_MAPPING
from .database import DataBase
from .ldm_classes import Filter, RequestDataObjectsReq


class TinyDB(DataBase):
    """
    Class inherting from DataBase.

    This class is used to store data in a TinyDB database. TinyDB is a lightweight document oriented database.
    """

    def __init__(self, database_name: str = None, database_path: str = None):
        """
        Parameters
        ----------
        path : str
            Path to the database file.
        """
        if database_path is None:
            raise ValueError("Database path not specified. Using TinyDB requires a path")
        self.database_path = database_path
        self.database_name = database_name
        self.database = tinydb.TinyDB(database_name)

    def delete(self, database_name: str = None) -> bool:
        """
        Delete database with a given name, returns a boolean stating if deletion has been succesful.

        Parameters
        ----------
        None
        """
        self.database.close()
        try:
            os.remove(self.database_path + self.database_name)
        except FileNotFoundError as e:
            print("File not found: " + str(e))
            return False
        return True

    def create_query_from_filter_statement(
        self, query: tinydb.Query, attribute: str
    ) -> tinydb.Query():
        """
        Method to create querry from filter statement.
        i.e. "cam.camParameters.basicContainer.stationType" == 1

        Parameters
        ----------
        query : Query -> TinyDB query object
        attribute : str -> attribute to be searched for
        """
        nested_fields = attribute.split(".")
        # Dynamically build the query
        for field in nested_fields:
            query = getattr(query, field)

        return query

    def create_query_search(
        self, query_with_attribute: tinydb.Query, operator: str, ref_value: int
    ) -> tinydb.Query:
        """
        Method to create querry from filter statement (i.e. "cam.camParameters.basicContainer.stationType" == 1)

        Parameters
        ----------

        """
        compare_function = OPERATOR_MAPPING.get(operator)
        if compare_function is not None:
            return compare_function(query_with_attribute, ref_value)

        raise ValueError(
            "Operator not supported according to ETSI TS 102 894-2 V2.2.1 (2023-10)"
        )

    def parse_filter_statement(self, query: tinydb.Query, filter: Filter) -> str:
        """
        Method to parse filter statement to the TinyDB query language.

        Parameters
        ----------
        query : Query -> TinyDB query object
        filter : Filter -> Filter to be parsed

        Returns
        -------
        str -> TinyDB query language
        """
        if filter.filter_statement_2 is not None:
            if str(filter.logical_operator) == "and":
                return self.create_query_search(
                    self.create_query_from_filter_statement(
                        query, filter.filter_statement_1.attribute
                    ),
                    str(filter.filter_statement_1.operator),
                    filter.filter_statement_1.ref_value,
                ) & self.create_query_search(
                    self.create_query_from_filter_statement(
                        query, filter.filter_statement_2.attribute
                    ),
                    str(filter.filter_statement_2.operator),
                    filter.filter_statement_2.ref_value,
                )
            return self.create_query_search(
                self.create_query_from_filter_statement(
                    query, filter.filter_statement_1.attribute
                ),
                str(filter.filter_statement_1.operator),
                filter.filter_statement_1.ref_value,
            ) | self.create_query_search(
                self.create_query_from_filter_statement(
                    query, filter.filter_statement_2.attribute
                ),
                str(filter.filter_statement_2.operator),
                filter.filter_statement_2.ref_value,
            )
        return self.create_query_search(
            self.create_query_from_filter_statement(
                query, filter.filter_statement_1.attribute
            ),
            str(filter.filter_statement_1.operator),
            filter.filter_statement_1.ref_value,
        )

    def search(self, data_request: RequestDataObjectsReq):
        """
        Search for data with a Filter (from ETSI ETSI EN 302 895 V1.1.1 (2014-09).

        Parameters
        ----------
        query : Filter
            Filter to be used for the search.
        """
        query = self.parse_filter_statement(tinydb.Query(), data_request.filter)
        return self.database.search(query)

    def insert(self, data: dict) -> int:
        """
        Insert data into the database, returns the index of the inserted data.

        Parameters
        ----------
        data : dict
            Data to be inserted into the database.
        """
        return self.database.insert(data)

    def get(self, index: int) -> list:
        """
        Get data from the database with a given index.

        Parameters
        ----------
        index : int
            Index of the data to be retrieved.
        """
        return self.database.get(doc_id=index)

    def update(self, data: dict, index: int) -> bool:
        """
        Update data in the database with a given index, returns a boolean stating if update has been succesful.

        Parameters
        ----------
        data : dict
            Data to be updated in the database.
        index : int
            Index of the data to be updated.
        """
        self.database.update(data, doc_ids=[index])
        return True

    def remove(self, data_object: dict = None) -> bool:
        """
        Remove data from the database with a given index, returns a boolean stating if removal has been succesful.

        Parameters
        ----------
        data_object : dict
            Data to be removed from the database.
        index : int
            Index of the data to be removed.
        """
        self.database.remove(doc_ids=[data_object.doc_id])
        return True

    def all(self) -> list:
        """
        Get all data from the database.

        Returns
        -------
        list
            All data from the database.
        """
        return self.database.all()

    def exists(self, field_name: str, data_object_id: int = None) -> bool:
        """
        Check if specific field exists

        Parameters
        ----------
        field_name : str
            Name of field to be checked
        Returns
        -------
        bool
            Indicates whether field exists
        """
        if field_name == "dataObjectID":
            return self.database.contains(doc_id=data_object_id)
        query = tinydb.Query()
        nested_fields = field_name.split(".")

        # Dynamically build the query
        for field in nested_fields:
            query = getattr(query, field)
        if self.search(query.exists()):
            return True
        return False



from __future__ import annotations
from collections.abc import Callable
import multiprocessing
import threading

# pylint: disable=import-error
from cv2xlinklayer import CV2XLinkLayer
from .link_layer import LinkLayer


class PythonCV2XLinkLayer(LinkLayer):
    """
    Link Layer to directly send via C-V2X. Uses the Telematics SDK from Qualcomm.
    This link Layer requires the usage of a pre-compiled library.

    Attributes
    ----------
    receive_callback : Callable[[bytes], None]
        Callback to call when a packet is received.
    link_layer : CV2XLinkLayer
        Link Layer to send and receive C-V2X messages.
    process : multiprocessing.Process
        Process to run the receive function.
    callback_thread : threading.Thread
        Thread to handle callbacks in the main memory space.
    callback_queue : multiprocessing.Queue
        Queue for transferring data between processes for callback execution.
    stop_event : multiprocessing.Event
        Event to signal the process to stop.
    """

    def __init__(self, receive_callback: Callable[[bytes], None]) -> None:
        super().__init__(receive_callback)
        self.link_layer = CV2XLinkLayer()
        self.stop_event = multiprocessing.Event()
        self.callback_queue = multiprocessing.Queue()

        # Process for receiving data
        self.process = multiprocessing.Process(
            target=self.receive_process,
            args=(self.callback_queue, self.stop_event),
            daemon=True,
        )
        self.process.start()

        # Thread for handling callbacks in the main process
        self.callback_thread = threading.Thread(
            target=self.callback_handler_loop, args=(self.callback_queue,), daemon=True
        )
        self.callback_thread.start()

    def send(self, packet: bytes) -> None:
        """
        Sends a packet via C-V2X.

        Parameters
        ----------
        packet : bytes
            Packet to send.
        """
        self.link_layer.send(b"\x03" + packet)

    def receive_process(
        self, callback_queue: multiprocessing.Queue, stop_event: multiprocessing.Event
    ) -> None:
        """
        Process to receive data from CV2XLinkLayer and place it in the callback queue.

        Parameters
        ----------
        callback_queue : multiprocessing.Queue
            Queue for transferring received data to the main process.
        stop_event : multiprocessing.Event
            Event to signal the process to stop.
        """
        while not stop_event.is_set():
            data = self.link_layer.receive()
            if data:
                data = data[1:]  # Remove the first byte as per protocol
                callback_queue.put(data)  # Place data in the queue

    def callback_handler_loop(self, callback_queue: multiprocessing.Queue) -> None:
        """
        Thread loop to handle callbacks from the callback queue.

        Parameters
        ----------
        callback_queue : multiprocessing.Queue
            Queue for transferring received data to the main process.
        """
        while True:
            data = callback_queue.get()
            if data is None:  # Stop signal
                break
            if self.receive_callback:
                self.receive_callback(data)

    def stop(self) -> None:
        """
        Stops the receive process, callback thread, and cleans up resources.
        """
        # Signal the process to stop
        self.stop_event.set()
        self.process.join()

        # Stop the callback thread
        self.callback_queue.put(None)  # Send a stop signal to the thread
        self.callback_thread.join()



from time import time
import logging

from ..facilities.local_dynamic_map.ldm_facility import LDMFacility
from ..facilities.local_dynamic_map.ldm_classes import (
    SubscribeDataObjectsResp,
    SubscribeDataobjectsReq,
    RequestDataObjectsResp,
    RegisterDataConsumerReq,
    RegisterDataConsumerResp,
)
from ..facilities.local_dynamic_map.ldm_constants import SPATEM, CAM, VAM, DENM
from ..facilities.ca_basic_service.cam_transmission_management import GenerationDeltaTime

from ..utils.static_location_service import ThreadStaticLocationService as Location

from .prometheus_adaptation import PrometheusClientPull


class MetricsExposer:
    def __init__(self, its_station_name: str, ldm: LDMFacility, location: Location) -> None:
        """
        Initialization class for MetricsExposer
        """
        self.logging = logging.getLogger("metrics")

        self.its_station_name = its_station_name
        self.ldm = ldm
        self.generation_delta_time = GenerationDeltaTime()
        self.prometheus = PrometheusClientPull()
        self.__register_to_ldm(ldm, location)
        self.__subscribe_to_ldm(ldm)

        self.logging.info("Metrics Exposer initialized!")

    def __register_to_ldm(self, ldm: LDMFacility, location: Location) -> None:
        """
        Register to the Local Data Manager (LDM) facility for sending the data.

        Parameters
        ----------
        ldm : LDMFacility
            Local Data Manager (LDM) facility object.

        Returns
        -------
        None
        """
        register_data_consumer_reponse: RegisterDataConsumerResp = ldm.if_ldm_4.register_data_consumer(
            RegisterDataConsumerReq(
                application_id=SPATEM,  # TODO: Allow application to sign up!!
                access_permisions=[VAM, CAM],
                area_of_interest=location,
            )
        )
        if register_data_consumer_reponse.result == 2:
            raise Exception(f"Failed to register data consumer: {str(register_data_consumer_reponse)}")
        self.logging.debug(f"Registered to LDM with response: {str(register_data_consumer_reponse)}")

    def __current_its_time(self) -> float:
        """
        Get the current ITS time.

        Returns
        -------
        float
            Current ITS time.
        """
        self.generation_delta_time.set_in_normal_timestamp(time() * 1000)
        return self.generation_delta_time.msec

    def __handle_cam_data_object(self, data_object: dict) -> None:
        """
        Handle the CAM data object received from the Local Data Manager (LDM).

        Parameters
        ----------
        data_object : DataObject
            Data object received from the LDM.

        Returns
        -------
        None
        """
        latency = self.__current_its_time() - data_object["timeStamp"]
        self.prometheus.send_latency(latency)
        self.prometheus.send_ldm_map(
            data_object["header"]["stationId"],
            data_object["cam"]["camParameters"]["basicContainer"]["stationType"],
            self.its_station_name,
            data_object["cam"]["camParameters"]["basicContainer"]["referencePosition"]["latitude"] / 10000000,
            data_object["cam"]["camParameters"]["basicContainer"]["referencePosition"]["longitude"] / 10000000,
        )
        self.logging.debug(
            f"Sending CAM latency and LDM map data to Prometheus, with values latency: {latency}, LDM map, "
            + f"lat: {data_object['cam']['camParameters']['basicContainer']['referencePosition']['latitude'] / 10000000}, "
            + f"lon: {data_object['cam']['camParameters']['basicContainer']['referencePosition']['longitude'] / 10000000}"
        )

    def __handle_vam_data_object(self, data_object: dict) -> None:
        """
        Handle the VAM data object received from the Local Data Manager (LDM).

        Parameters
        ----------
        data_object : DataObject
            Data object received from the LDM.

        Returns
        -------
        None
        """
        latency = self.__current_its_time() - data_object["timeStamp"]
        self.prometheus.send_latency(latency)
        self.prometheus.send_ldm_map(
            data_object["dataObject"]["header"]["stationId"],
            data_object["dataObject"]["vam"]["vamParameters"]["basicContainer"]["stationType"],
            self.its_station_name,
            data_object["dataObject"]["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["latitude"]
            / 10000000,
            data_object["dataObject"]["vam"]["vamParameters"]["basicContainer"]["referencePosition"]["longitude"]
            / 10000000,
        )

        self.logging.debug(
            f"Sending VAM latency and LDM map data to Prometheus, with values latency: {latency}, LDM map, "
            + f"lat: {data_object['dataObject']['vam']['vamParameters']['basicContainer']['referencePosition']['latitude'] / 10000000}, "
            + f"lon: {data_object['dataObject']['vam']['vamParameters']['basicContainer']['referencePosition']['longitude'] / 10000000}"
        )

    def __handle_data_object(self, data_object: dict) -> None:
        """
        Handle the data object received from the Local Data Manager (LDM).

        Parameters
        ----------
        data_object : DataObject
            Data object received from the LDM.

        Returns
        -------
        None
        """
        if data_object["application_id"] == CAM:
            self.__handle_cam_data_object(data_object)
        elif data_object["application_id"] == VAM:
            self.__handle_vam_data_object(data_object)
        else:
            print(f"Unknown data object type: {data_object.data_object}")

    def __ldm_subscription_callback(self, data_object: RequestDataObjectsResp) -> None:
        """
        Callback function for receiving the data objects from the Local Data Manager (LDM).

        Parameters
        ----------
        data_object : DataObject
            Data object received from the LDM.

        Returns
        -------
        None
        """
        if data_object.application_id == SPATEM:
            for data in data_object.data_objects:
                self.__handle_data_object(data)
        else:
            raise Exception(f"Unknown data object type: {data_object.data_object_type}")

    def __subscribe_to_ldm(self, ldm: LDMFacility) -> None:
        """
        Subscribe to the Local Data Manager (LDM) facility for receiving the data.

        Parameters
        ----------
        ldm : LDMFacility
            Local Data Manager (LDM) facility object.

        Returns
        -------
        None
        """
        subscribe_data_consumer_response: SubscribeDataObjectsResp = ldm.if_ldm_4.subscribe_data_consumer(
            SubscribeDataobjectsReq(
                application_id=SPATEM,
                data_object_type=[CAM, VAM, DENM],
                priority=None,
                filter=None,
                notify_time=0.5,
                multiplicity=None,
                order=None,
            ),
            self.__ldm_subscription_callback,
        )
        if subscribe_data_consumer_response.result.result != 0:
            raise Exception(f"Failed to subscribe to data objects: {str(subscribe_data_consumer_response.result)}")

        self.logging.debug(f"Subscribed to LDM with response: {str(subscribe_data_consumer_response)}")



from dateutil import parser

from .exceptions import DecodeError
from .gn_address import GNAddress


class TST:
    """
    Timestamp class.  ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.5.2

    Expresses the time in milliseconds at which
    the latitude and longitude of the ITS-S were
    acquired by the GeoAdhoc router. The time is
    encoded as:
    TST = TST(TAI)mod 2^32
    where TST(TAI) is the number of elapsed TAI
    milliseconds since 2004-01-01 00:00:00.000
    UTC

    Attributes
    ----------
    msec : int
        Milliseconds.
    """

    def __init__(self) -> None:
        self.msec = 0

    def set_in_normal_timestamp_seconds(self, timestamp: int) -> None:
        """
        Set the timestamp in normal timestamp format.

        Parameters
        ----------
        timestamp : int
            Timestamp in normal timestamp format.
        """
        self.msec = ((timestamp-1072911600)*1000) % 2**32

    def set_in_normal_timestamp_milliseconds(self, timestamp: int) -> None:
        """
        Set the timestamp in normal timestamp format.

        Parameters
        ----------
        timestamp : int
            Timestamp in normal timestamp format.
        """
        self.msec = (timestamp-1072911600000) % 2**32

    def encode(self) -> int:
        """
        Encode the timestamp.

        Returns
        -------
        int
            Encoded timestamp.
        """
        return self.msec % 2**32

    def decode(self, data: int) -> None:
        """
        Decode the timestamp.

        Parameters
        ----------
        data : int
            Encoded timestamp.
        """
        self.msec = data % 2**32

    def __eq__(self, __o: object) -> bool:
        """
        Equal operator.

        Parameters
        ----------
        __o : object
            Object to compare.

        Returns
        -------
        bool
            True if equal, False otherwise.
        """
        if isinstance(__o, TST):
            return self.msec == __o.msec
        return False

    def __ne__(self, __o: object) -> bool:
        """
        Not equal operator.

        Parameters
        ----------
        __o : object
            Object to compare.

        Returns
        -------
        bool
            True if not equal, False otherwise.
        """
        return not self.__eq__(__o)

    def __gt__(self, __o: object) -> bool:
        """
        Greater than operator.
        Uses the ETSI EN 302 636-4-1 V1.4.1 (2020-01). Annex C.2

        Parameters
        ----------
        __o : object
            Object to compare.

        Returns
        -------
        bool
            True if greater than, False otherwise.
        """
        if isinstance(__o, TST):
            return (self.msec > __o.msec) and ((self.msec - __o.msec) <= (2**32)/2) or (
                (__o.msec > self.msec) and ((__o.msec - self.msec) > (2**32)/2))
        return False

    def __ge__(self, __o: object) -> bool:
        """
        Greater than or equal operator.

        Parameters
        ----------
        __o : object
            Object to compare.

        Returns
        -------
        bool
            True if greater than or equal, False otherwise.
        """
        if isinstance(__o, TST):
            return self.__eq__(__o) or self.__gt__(__o)
            # return (self==__o) or (self>__o)
        return False

    def __lt__(self, __o: object) -> bool:
        """
        Less than operator.

        Parameters
        ----------
        __o : object
            Object to compare.

        Returns
        -------
        bool
            True if less than, False otherwise.
        """
        if isinstance(__o, TST):
            # return not (self>=__o)
            return not self.__ge__(__o)
        return False

    def __le__(self, __o: object) -> bool:
        """
        Less than or equal operator.

        Parameters
        ----------
        __o : object
            Object to compare.

        Returns
        -------
        bool
            True if less than or equal, False otherwise.
        """
        if isinstance(__o, TST):
            # return not (self>__o)
            return not self.__gt__(__o)
        return False

    def __add__(self, __o: object) -> int:
        """
        Add operator.

        Parameters
        ----------
        __o : object
            Object to add.

        Returns
        -------
        int
            Sum.
        """
        if isinstance(__o, TST):
            return (self.msec + __o.msec) % 2**32
        return self.msec

    def __sub__(self, __o: object) -> int:
        """
        Subtract operator.

        Parameters
        ----------
        __o : object
            Object to subtract.

        Returns
        -------
        int
            Difference.
        """
        if isinstance(__o, TST):
            result = self.msec - __o.msec
            if result < 0:
                result += 2**32
            return result
        return self.msec

    def __str__(self) -> str:
        """
        String representation.

        Returns
        -------
        str
            String representation.
        """
        return str(self.msec)+" msec"


class LongPositionVector:
    """
    Long Position Vector class.  ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.5.2

    Attributes
    ----------
    gn_addr : GNAddress
        GeoNetworking address GN_ADDR.
    tst : TST
        Timestamp TST indicating when the geographical position POS was generated.
    Latitude : int
        Latitude of the ITS-S. WGS 84 1/10 micro degree.
    Longitude : int
        Longitude of the ITS-S. WGS 84 1/10 micro degree.
    pai : bool
        Accuracy of the geographical position PAI. Set to 1 (i.e. True) if the semiMajorConfidence
        of the PosConfidenceEllipse as specified in ETSI TS 102 894-2 [11] is smaller than the GN
        protocol constant itsGnPaiInterval / 2 Set to 0 (i.e. False) otherwise
    s : int
        Speed S. 0.01 m/s.
    h : int
        Heading H. 0.1 degree from north.
    """

    def __init__(self) -> None:
        self.gn_addr = GNAddress()
        self.tst = TST()
        self.latitude = 0
        self.longitude = 0
        self.pai = False
        self.s = 0
        self.h = 0

    def set_gn_addr(self, gn_addr: GNAddress) -> None:
        """
        Set the GN address.

        Parameters
        ----------
        gn_addr : GNAddress
            GN address.
        """
        self.gn_addr = gn_addr

    def set_tst_in_normal_timestamp_seconds(self, timestamp: int) -> None:
        """
        Set the timestamp in normal timestamp format.

        Parameters
        ----------
        timestamp : int
            Timestamp in normal timestamp format.
        """
        self.tst.set_in_normal_timestamp_seconds(timestamp)

    def set_tst_in_normal_timestamp_milliseconds(self, timestamp: int) -> None:
        """
        Set the timestamp in normal timestamp format.

        Parameters
        ----------
        timestamp : int
            Timestamp in normal timestamp format.
        """
        self.tst.set_in_normal_timestamp_milliseconds(timestamp)

    def set_latitude(self, latitude: float) -> None:
        """
        Set the latitude.

        Parameters
        ----------
        latitude : float
            Latitude in WGS 84 1/10 micro degree.
        """
        self.latitude = int(latitude*10000000)

    def set_longitude(self, longitude: float) -> None:
        """
        Set the longitude.

        Parameters
        ----------
        longitude : float
            Longitude in WGS 84 1/10 micro degree.
        """
        self.longitude = int(longitude*10000000)

    def set_pai(self, pai: bool) -> None:
        """
        Set the PAI.

        Parameters
        ----------
        pai : bool
            PAI.
        """
        self.pai = pai

    def set_speed(self, speed: float) -> None:
        """
        Set the speed.

        Parameters
        ----------
        speed : float
            Speed in m/s.
        """
        self.s = int(speed*100)

    def set_heading(self, heading: float) -> None:
        """
        Set the heading.

        Parameters
        ----------
        heading : float
            Heading in degree from north.
        """
        self.h = int(heading*10)

    def refresh_with_tpv_data(self, tpv_data: dict) -> None:
        """
        Refresh the LongPositionVector with a dict containing the data from a GPSD TPV message.

        Parameters
        ----------
        tpv_data : dict
            Dict containing the data from a GPSD TPV message.
        """
        self.set_tst_in_normal_timestamp_seconds(
            int(parser.parse(tpv_data["time"]).timestamp()))
        self.set_latitude(tpv_data['lat'])
        self.set_longitude(tpv_data['lon'])
        self.set_speed(tpv_data['speed'])
        self.set_heading(tpv_data['track'])

    def encode(self) -> bytes:
        """
        Encode the LongPositionVector.

        Returns
        -------
        bytes
            Encoded LongPositionVector.
        """
        return ((self.gn_addr.encode_to_int() << 32*4) | (self.tst.encode() << 32*3) | (
            self.latitude << 32*2) | (self.longitude << 32) | (
                self.pai << 31) | (self.s << 16) | self.h).to_bytes(24, byteorder='big')

    def encode_to_int(self) -> int:
        """
        Encode the LongPositionVector.

        Returns
        -------
        int
            Encoded LongPositionVector.
        """
        return (self.gn_addr.encode_to_int() << 32*4) | (self.tst.encode() << 32*3) | (
            self.latitude << 32*2) | (self.longitude << 32) | (self.pai << 31) | (self.s << 16) | self.h

    def decode(self, data: bytes) -> None:
        """
        Decode the LongPositionVector.

        Parameters
        ----------
        data : bytes
            Encoded LongPositionVector.
        """
        if len(data) < 24:
            raise DecodeError("LongPositionVector must be 24 bytes long")
        data_as_int = int.from_bytes(data[0:24], byteorder='big')
        self.gn_addr.decode((data_as_int >> 32*4).to_bytes(8, byteorder='big'))
        self.tst.decode(data_as_int >> 32*3)
        self.latitude = (data_as_int >> 32*2) & 0xFFFFFFFF
        self.longitude = (data_as_int >> 32) & 0xFFFFFFFF
        self.pai = (data_as_int >> 31) & 0x1
        self.s = (data_as_int >> 16) & 0x7FFF
        self.h = data_as_int & 0xFFFF

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, LongPositionVector):
            return self.gn_addr == __o.gn_addr and self.tst == __o.tst and self.latitude == __o.latitude and (
                self.longitude == __o.longitude) and self.pai == __o.pai and self.s == __o.s and self.h == __o.h
        return False

    def __str__(self):
        return "LongPositionVector: \n" + \
            "    GN Address: " + str(self.gn_addr) + "\n" + \
            "    Timestamp: " + str(self.tst) + "\n" + \
            "    Latitude: " + str(self.latitude) + "\n" + \
            "    Longitude: " + str(self.longitude) + "\n" + \
            "    PAI: " + str(self.pai) + "\n" + \
            "    Speed: " + str(self.s) + "\n" + \
            "    Heading: " + str(self.h) + "\n"


class ShortPositionVector:
    """
    Short Position Vector class.  ETSI EN 302 636-4-1 V1.4.1 (2020-01). Section 9.5.2

    Attributes
    ----------
    gn_addr : GNAddress
        GeoNetworking address GN_ADDR.
    tst : TST
        Timestamp TST indicating when the geographical position POS was generated.
    latitude : int
        Latitude of the ITS-S. WGS 84 1/10 micro degree.
    longitude : int
        Longitude of the ITS-S. WGS 84 1/10 micro degree.
    """

    def __init__(self) -> None:
        self.gn_addr = GNAddress()
        self.tst = TST()
        self.latitude = 0
        self.longitude = 0

    def set_gn_addr(self, gn_addr: GNAddress) -> None:
        """
        Set the GN address.

        Parameters
        ----------
        gn_addr : GNAddress
            GN address.
        """
        self.gn_addr = gn_addr

    def set_tst_in_normal_timestamp_seconds(self, timestamp: int) -> None:
        """
        Set the timestamp in normal timestamp format.

        Parameters
        ----------
        timestamp : int
            Timestamp in normal timestamp format.
        """
        self.tst.set_in_normal_timestamp_seconds(timestamp)

    def set_tst_in_normal_timestamp_milliseconds(self, timestamp: int) -> None:
        """
        Set the timestamp in normal timestamp format.

        Parameters
        ----------
        timestamp : int
            Timestamp in normal timestamp format.
        """
        self.tst.set_in_normal_timestamp_milliseconds(timestamp)

    def set_latitude(self, latitude: float) -> None:
        """
        Set the latitude.

        Parameters
        ----------
        latitude : float
            Latitude in WGS 84 1/10 micro degree.
        """
        self.latitude = int(latitude*10000000)

    def get_latitude(self) -> float:
        """
        Get the latitude.

        Returns
        -------
        float
            Latitude in WGS 84 1/10 micro degree.
        """
        return self.latitude/10000000

    def set_longitude(self, longitude: float) -> None:
        """
        Set the longitude.

        Parameters
        ----------
        longitude : float
            Longitude in WGS 84 1/10 micro degree.
        """
        self.longitude = int(longitude * 10000000)

    def get_longitude(self) -> float:
        """
        Get the longitude.

        Returns
        -------
        float
            Longitude in WGS 84 1/10 micro degree.
        """
        return self.longitude/10000000

    def encode(self) -> bytes:
        """
        Encode the ShortPositionVector.

        Returns
        -------
        bytes
            Encoded ShortPositionVector.
        """
        return ((self.gn_addr.encode_to_int() << 32*3) | (self.tst.encode() << 32*2) | (
            self.latitude << 32*1) | self.longitude).to_bytes(20, byteorder='big')

    def encode_to_int(self) -> int:
        """
        Encode the ShortPositionVector.

        Returns
        -------
        int
            Encoded ShortPositionVector.
        """
        return (self.gn_addr.encode_to_int() << 32*3) | (self.tst.encode() << 32*2) | (
            self.latitude << 32*1) | self.longitude

    def decode(self, data: bytes) -> None:
        """
        Decode the ShortPositionVector.

        Parameters
        ----------
        data : bytes
            Encoded ShortPositionVector.
        """
        data = int.from_bytes(data, byteorder='big')
        self.gn_addr.decode((data >> 32*3).to_bytes(8, byteorder='big'))
        self.tst.decode(data >> 32*2)
        self.latitude = (data >> 32 * 1) & 0xFFFFFFFF
        self.longitude = data & 0xFFFFFFFF

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, ShortPositionVector):
            return self.gn_addr == __o.gn_addr and self.tst == __o.tst and (
                self.latitude == __o.latitude) and self.longitude == __o.longitude
        return False
